/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE accset.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/08/11 09:02:29 Alf*/

#include "log.h"
#include "accset.h"

#include <stdlib.h>

#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

accset_t *g_set = NULL;

static GtkWidget *set_net_new (accset_t *);

static void set_save (GtkWidget *, accset_t *);
static void set_cancel (GtkWidget *, accset_t *);

#define load_int_set(token,value,min,max,def)   do {            \
    if (dic != NULL)                                            \
      {                                                         \
        intv = iniparser_getint (dic, (token), (min));          \
        (value) = intv > (min) && intv < (max) ? intv : (def);  \
      }                                                         \
    else                                                        \
      {                                                         \
        (value) = (def);                                        \
      }                                                         \
} while (0)

#define load_str_set(token,value,def)               do {        \
    if (dic != NULL)                                            \
      {                                                         \
        strv = iniparser_getstring (dic, (token), NULL);        \
        (value) = strv == NULL ? strdup (def) : strdup (strv);  \
      }                                                         \
    else                                                        \
      {                                                         \
        (value) = strdup (def);                                 \
      }                                                         \
} while (0)

accset_t *
accset_new (const char *cfile)
{
  accset_t *set;
  char *strv;
  int intv;
  GtkWidget *notebook, *hbox, *bsave, *bcancel, *wnet;
  dictionary *dic;

  set = (accset_t *) calloc (1, sizeof (accset_t));
  if (set == NULL)
    {
      acc_err (_("Memory error\n"));
      return NULL;
    }

  set->widget = gtk_vbox_new (FALSE, 2);

  notebook = gtk_notebook_new ();
  gtk_box_pack_start (GTK_BOX (set->widget), notebook, TRUE, TRUE, 2);

  hbox = gtk_hbox_new (TRUE, 5);
  gtk_box_pack_start (GTK_BOX (set->widget), hbox, FALSE, FALSE, 2);

  bsave = gtk_button_new_with_label (_("Save"));
  gtk_box_pack_start (GTK_BOX (hbox), bsave, TRUE, FALSE, 2);
  g_signal_connect (G_OBJECT (bsave), "clicked", G_CALLBACK (set_save), set);
  bcancel = gtk_button_new_with_label (_("Cancel"));
  gtk_box_pack_start (GTK_BOX (hbox), bcancel, TRUE, FALSE, 2);
  g_signal_connect (G_OBJECT (bcancel), "clicked", G_CALLBACK (set_cancel),
                    set);

  strv = acc_realpath (cfile);
  if (strv != NULL)
    {
      set->file = g_locale_from_utf8 (strv, -1, NULL, NULL, NULL);
      free (strv);
    }

  if (set->file)
    {
      dic = set->dic = iniparser_load (set->file);
#ifdef _DEBUG
      iniparser_dump (set->dic, stderr);
#endif
    }
  else
    {
      dic = NULL;
    }

  load_int_set ("pc:level", set->pc_level, 0, 8, 3);

#ifdef WIN32
  load_str_set ("pc:engine", set->pc_engine, "acnguy.exe");
#else
  load_str_set ("pc:engine", set->pc_engine, CMAKE_INSTALL_PREFIX"/bin/acnguy");
#endif

  load_int_set ("net:conn_timeout", set->net_conn_timeout, 0, 10, 4);
  load_int_set ("net:read_timeout", set->net_read_timeout, 0, 6, 2);

  load_int_set ("net:refresh_user_timer", set->net_ref_user, 300, 3600, 600);

  load_int_set ("net:max_users", set->net_max_users, 0, 60, 1000);

  load_int_set ("net:ip", set->net_ip, 0, INT_MAX, 0);
  load_int_set ("net:tcp_port", set->net_tcp, 2000, 65534, 23456);
  load_int_set ("net:udp_port", set->net_udp, 2000, 65534, 23456);

  wnet = set_net_new (set);
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), wnet,
                            gtk_label_new (_("Net")));
  gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), 0);

  if (g_set == NULL)
    {
      g_set = set;
    }

  return set;
}

void
accset_delete (accset_t * set)
{
  if (set->file)
    {
      if (set->dic)
        {
          FILE *fp;

          fp = fopen (set->file, "w");
          if (fp)
            {
              iniparser_dump_ini (set->dic, fp);
              fclose (fp);
            }
        }

      g_free (set->file);
    }

  if (set->dic)
    {
      iniparser_freedict (set->dic);
    }

  free (set);
}

static GtkWidget *
set_net_new (accset_t * set)
{
  GtkWidget *widget, *hbox, *label;
  char buf[100];

  widget = gtk_vbox_new (FALSE, 2);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (widget), hbox, FALSE, FALSE, 2);
  label = gtk_label_new (_("Conn timeout"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  set->wid_net_conn_timeout = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), set->wid_net_conn_timeout, FALSE, FALSE,
                      2);
  g_snprintf (buf, sizeof buf, "%d", set->net_conn_timeout);
  gtk_entry_set_text (GTK_ENTRY (set->wid_net_conn_timeout), buf);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (widget), hbox, FALSE, FALSE, 2);
  label = gtk_label_new (_("Read timeout"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  set->wid_net_read_timeout = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), set->wid_net_read_timeout, FALSE, FALSE,
                      2);
  g_snprintf (buf, sizeof buf, "%d", set->net_read_timeout);
  gtk_entry_set_text (GTK_ENTRY (set->wid_net_read_timeout), buf);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (widget), hbox, FALSE, FALSE, 2);
  label = gtk_label_new (_("Refresh User Timer"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  set->wid_net_ref_user = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), set->wid_net_ref_user, FALSE, FALSE, 2);
  g_snprintf (buf, sizeof buf, "%d", set->net_ref_user);
  gtk_entry_set_text (GTK_ENTRY (set->wid_net_ref_user), buf);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (widget), hbox, FALSE, FALSE, 2);
  label = gtk_label_new (_("Refresh User Timer"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  set->wid_net_max_users = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), set->wid_net_max_users, FALSE, FALSE,
                      2);
  g_snprintf (buf, sizeof buf, "%d", set->net_max_users);
  gtk_entry_set_text (GTK_ENTRY (set->wid_net_max_users), buf);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (widget), hbox, FALSE, FALSE, 2);
  label = gtk_label_new (_("Local IP"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  set->wid_net_ip = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), set->wid_net_ip, FALSE, FALSE, 2);
  g_snprintf (buf, sizeof buf, "%s",
              inet_ntoa (*(struct in_addr *) (&set->net_ip)));
  gtk_entry_set_text (GTK_ENTRY (set->wid_net_ip), buf);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (widget), hbox, FALSE, FALSE, 2);
  label = gtk_label_new (_("TCP Port"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  set->wid_net_tcp = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), set->wid_net_tcp, FALSE, FALSE, 2);
  g_snprintf (buf, sizeof buf, "%d", set->net_tcp);
  gtk_entry_set_text (GTK_ENTRY (set->wid_net_tcp), buf);

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_box_pack_start (GTK_BOX (widget), hbox, FALSE, FALSE, 2);
  label = gtk_label_new (_("UDP Port"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  set->wid_net_udp = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), set->wid_net_udp, FALSE, FALSE, 2);
  g_snprintf (buf, sizeof buf, "%d", set->net_udp);
  gtk_entry_set_text (GTK_ENTRY (set->wid_net_udp), buf);

  return widget;
}

extern void iniparser_set (dictionary *, const char *, const char *);
static void
set_save (GtkWidget * wid, accset_t * set)
{
  const gchar *string;
  char buf[0x100];
  FILE *fp;

  g_snprintf (buf, sizeof buf, "%d", set->pc_level);
  iniparser_set (set->dic, "pc:level", buf);

  g_snprintf (buf, sizeof buf, "%s", set->pc_engine);
  iniparser_set (set->dic, "pc:engine", buf);

  string = gtk_entry_get_text (GTK_ENTRY (set->wid_net_conn_timeout));
  iniparser_set (set->dic, "net:conn_timeout", string);

  string = gtk_entry_get_text (GTK_ENTRY (set->wid_net_read_timeout));
  iniparser_set (set->dic, "net:read_timeout", string);

  string = gtk_entry_get_text (GTK_ENTRY (set->wid_net_ref_user));
  iniparser_set (set->dic, "net:refresh_user_timer", string);

  string = gtk_entry_get_text (GTK_ENTRY (set->wid_net_max_users));
  iniparser_set (set->dic, "net:max_users", string);

  string = gtk_entry_get_text (GTK_ENTRY (set->wid_net_ip));
  snprintf (buf, sizeof buf, "%u", inet_addr (string));
  iniparser_set (set->dic, "net:ip", buf);

  string = gtk_entry_get_text (GTK_ENTRY (set->wid_net_tcp));
  iniparser_set (set->dic, "net:tcp_port", string);

  string = gtk_entry_get_text (GTK_ENTRY (set->wid_net_udp));
  iniparser_set (set->dic, "net:udp_port", string);

  fp = fopen (set->file, "w");
  if (fp)
    {
      iniparser_dump_ini (set->dic, fp);
      fclose (fp);
    }
}

static void
set_cancel (GtkWidget * wid, accset_t * set)
{
}
