/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE accui.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/07/30 09:13:23 Alf*/

#include "log.h"
#include "accui.h"

#include <string.h>
#include <stdlib.h>

#define ACC_WIDTH               900
#define ACC_TOOLBAR_HEIGHT      50
#define ACC_STATUS_HEIGHT       25

static void ui_destroy_cb (GtkWidget *, accui_t *);

static void ui_main_cb (GtkWidget *, accui_t *);
static void ui_user_cb (GtkWidget *, accui_t *);
static void ui_setting_cb (GtkWidget *, accui_t *);

static GtkWidget *ui_toolbar_new (accui_t *);

static GtkWidget *ui_layout_new (accui_t *);

static GtkWidget *ui_status_new (accui_t *);

accui_t *g_ui = NULL;

accui_t *
accui_new (int *arc, char ***arv)
{
  accui_t *ui;
  GtkWidget *vbox;

  ui = calloc (1, sizeof (accui_t));
  if (ui == NULL)
    {
      acc_err (_("Memory error\n"));
      return NULL;
    }

  if (!g_thread_supported ())
    {
      g_thread_init (NULL);
    }
  gdk_threads_init ();

  gtk_init (arc, arv);

  ui->rootwindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (ui->rootwindow), PACKAGE_STRING);
  vbox = gtk_vbox_new (FALSE, 2);
  gtk_container_add (GTK_CONTAINER (ui->rootwindow), vbox);

  ui->set = accset_new ("~/.acnchess/config");
  ui->us = accus_new ();
  ui->xq = accxq_new ();

  ui->toolbar = ui_toolbar_new (ui);
  gtk_box_pack_start (GTK_BOX (vbox), ui->toolbar, FALSE, FALSE, 2);
  ui->layout = ui_layout_new (ui);
  gtk_box_pack_start (GTK_BOX (vbox), ui->layout, TRUE, TRUE, 2);
  ui->status = ui_status_new (ui);
  gtk_box_pack_end (GTK_BOX (vbox), ui->status, FALSE, FALSE, 2);

  gtk_window_set_default_size (GTK_WINDOW (ui->rootwindow), ACC_WIDTH,
                               ACC_WIDTH - (ACC_WIDTH >> 2));

  gtk_widget_show_all (ui->rootwindow);

  g_signal_connect (G_OBJECT (ui->rootwindow), "destroy",
                    G_CALLBACK (ui_destroy_cb), ui);

  if (g_ui == NULL)
    {
      g_ui = ui;
    }

  return ui;
}

void
accui_run (accui_t * ui)
{
  gdk_threads_enter ();
  gtk_main ();
  gdk_threads_leave ();
}

static void
ui_destroy_cb (GtkWidget * wid, accui_t * ui)
{
  g_object_unref (ui->mlayout);
  g_object_unref (ui->ulayout);
  g_object_unref (ui->slayout);

  accus_delete (ui->us);
  accxq_delete (ui->xq);
  accset_delete (ui->set);

  gtk_main_quit ();

  free (ui);
}

static GtkWidget *
ui_toolbar_new (accui_t * ui)
{
  GtkWidget *toolbar, *bmain, *buser, *bsetting;

  toolbar = gtk_toolbar_new ();
  gtk_widget_set_size_request (toolbar, -1, ACC_TOOLBAR_HEIGHT);

  bmain = gtk_button_new_with_label (_("Main"));
  gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar), bmain, "main", "main");
  g_signal_connect (G_OBJECT (bmain), "clicked", G_CALLBACK (ui_main_cb), ui);

  buser = gtk_button_new_with_label (_("User"));
  gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar), buser, "user", "user");
  g_signal_connect (G_OBJECT (buser), "clicked", G_CALLBACK (ui_user_cb), ui);

  bsetting = gtk_button_new_with_label (_("Setting"));
  gtk_toolbar_append_widget (GTK_TOOLBAR (toolbar), bsetting, "setting",
                             "setting");
  g_signal_connect (G_OBJECT (bsetting), "clicked",
                    G_CALLBACK (ui_setting_cb), ui);

  return toolbar;
}

static GtkWidget *
ui_layout_new (accui_t * ui)
{
  GtkWidget *layout;

  layout = gtk_alignment_new ((gfloat) 0.01, (gfloat) 0.01, (gfloat) 1.0, (gfloat) 1.0);

  ui->mlayout = accxq_widget (ui->xq);
  g_object_ref (ui->mlayout);

  ui->ulayout = accus_widget (ui->us);
  g_object_ref (ui->ulayout);

  ui->slayout = accset_widget (ui->set);
  g_object_ref (ui->slayout);

  gtk_container_add (GTK_CONTAINER (layout), ui->mlayout);
  ui->current = ui->mlayout;

  return layout;
}

static int
ui_status_display (const char *s, void *arg)
{
  accui_t *ui;
  static guint id = 0;

  ui = (accui_t *) arg;

  if (id == 0)
    {
      id =
        gtk_statusbar_get_context_id (GTK_STATUSBAR (ui->status), "status");
    }
  else
    {
      gtk_statusbar_pop (GTK_STATUSBAR (ui->status), id);
    }

  gtk_statusbar_push (GTK_STATUSBAR (ui->status), id, s);

  return FALSE;
}

static GtkWidget *
ui_status_new (accui_t * ui)
{
  GtkWidget *status;

  status = gtk_statusbar_new ();
  gtk_widget_set_size_request (status, -1, ACC_STATUS_HEIGHT);

  acclog_register (ui_status_display, ui);

  return status;
}

static void
ui_main_cb (GtkWidget * wid, accui_t * ui)
{
  if (ui->current != NULL)
    {
      gtk_container_remove (GTK_CONTAINER (ui->layout), ui->current);
    }
  gtk_container_add (GTK_CONTAINER (ui->layout), ui->mlayout);
  ui->current = ui->mlayout;
  gtk_widget_show_all (ui->layout);
}

static void
ui_user_cb (GtkWidget * wid, accui_t * ui)
{
  if (ui->current != NULL)
    {
      gtk_container_remove (GTK_CONTAINER (ui->layout), ui->current);
    }
  gtk_container_add (GTK_CONTAINER (ui->layout), ui->ulayout);
  ui->current = ui->ulayout;
  gtk_widget_show_all (ui->layout);
}

static void
ui_setting_cb (GtkWidget * wid, accui_t * ui)
{
  if (ui->current != NULL)
    {
      gtk_container_remove (GTK_CONTAINER (ui->layout), ui->current);
    }
  gtk_container_add (GTK_CONTAINER (ui->layout), ui->slayout);
  ui->current = ui->slayout;
  gtk_widget_show_all (ui->layout);
}
