/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE accxq.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/07/30 13:43:21 Alf*/

#include "log.h"
#include "acnkern.h"
#include "qipu.h"
#include "accset.h"
#include "accxq.h"
#include "zgxq.h"
#include "accecco.h"

#include <stdlib.h>
#include <string.h>

#define PAN_CON_WIDTH           120
#define PAN_BUTTON_WIDTH        100
#define PAN_BUTTON_HEIGHT       30

extern accset_t *g_set;

accxq_t *g_xq = NULL;

enum pan_other_type { OTHER_PC = 0, OTHER_MANUAL };

/*
 * This functions's parameter are postion:
 * x and y
 */
static void pn_draw_zi_at (struct pan_node *, gint, gint, gint);
static void pn_draw_text_at (struct pan_node *, gint, gint, const gchar *);

static void pn_refresh (struct pan_node *);
static void pn_move (struct pan_node *, gint, gint, gint, gint);
static void pn_unmove (struct pan_node *);

static void pn_redraw_cb (GtkWidget *, GdkEventExpose *,
                          struct pan_node *);
static gboolean pn_press_cb (GtkWidget *, GdkEventButton *,
                             struct pan_node *);
static gboolean pn_release_cb (GtkWidget *, GdkEventButton *,
                               struct pan_node *);

static void qp_load_cb (GtkWidget *, struct pan_node *);
static void qp_save_cb (GtkWidget *, struct pan_node *);
static void qp_start_cb (GtkWidget *, struct pan_node *);
static void qp_branch_cb (GtkWidget *, struct pan_node *);
static void qp_next_cb (GtkWidget *, struct pan_node *);

static void qp_row_activated_cb (GtkTreeView *, GtkTreePath *, GtkTreeViewColumn *, struct pan_node *);

static void accxq_onload (accxq_t *);

static void pn_pcset_engine_select (GtkWidget *, struct pan_node *);

static void pan_baiqi_cb (GtkWidget *, struct pan_node *);
static void pan_start_cb (GtkWidget *, struct pan_node *);
static void pan_hui_cb (GtkWidget *, struct pan_node *);
static void pan_stop_cb (GtkWidget *, struct pan_node *);
static void pan_quit_cb (GtkWidget *, struct pan_node *);

static GtkWidget * pan_other_new (struct pan_node *);
static GtkWidget * pan_myself_new (struct pan_node *);
static GtkWidget * pan_draw_ui (accxq_t *, struct pan_node *);
static GtkWidget * pan_control_ui (accxq_t *, struct pan_node *);

static GtkWidget * pn_pcset_new (struct pan_node *);
static GtkWidget * pn_qpset_new (struct pan_node *);

static GdkPixbuf *xq_qi_buf (int, gint, gint);

static int xq_append_pan (accxq_t *, struct pan_node *);

accxq_t *
accxq_new ()
{
  accxq_t *xq;

  xq = calloc (1, sizeof (accxq_t));
  if (xq == NULL)
    {
      acc_err (_("Memory error\n"));
      return NULL;
    }

  ecco_init ();

  xq->widget = gtk_hbox_new (FALSE, 2);

  xq->pantab = gtk_notebook_new ();
  gtk_notebook_set_show_tabs (GTK_NOTEBOOK (xq->pantab), TRUE);
  gtk_notebook_set_scrollable (GTK_NOTEBOOK (xq->pantab), TRUE);
  gtk_notebook_set_show_border (GTK_NOTEBOOK (xq->pantab), TRUE);
  gtk_box_pack_start (GTK_BOX (xq->widget), xq->pantab, TRUE, TRUE, 2);

  accxq_onload (xq);

  if (g_xq == NULL)
    {
      g_xq = xq;
    }

  return xq;
}

struct pan_node *
pannode_new (accxq_t * xq, session_t * ses)
{
  struct pan_node *pn;
  GtkWidget *gdraw, *gcon;

  pn = calloc (1, sizeof (struct pan_node));
  if (pn == NULL)
    {
      acc_err (_("Memory error\n"));
      return NULL;
    }

  pn->session = ses;

  pn->state = PAN_STATE_TURN;

  pn->qipu = qipu_new ();

  pn->widget = gtk_hbox_new (FALSE, 1);
  gdraw = pan_draw_ui (xq, pn);
  gtk_box_pack_start (GTK_BOX (pn->widget), gdraw, TRUE, TRUE, 1);
  gcon = pan_control_ui (xq, pn);
  gtk_box_pack_end (GTK_BOX (pn->widget), gcon, FALSE, FALSE, 1);

  gtk_widget_show_all (pn->widget);

  pn->name = g_strdup (session_desc (ses));
  pn->accxq = xq;
  xq_append_pan (xq, pn);

  return pn;
}

void
pannode_delete (struct pan_node *xq)
{
  gint id;

  acc_debug ("delete accxq: %p\n", xq);

  if (xq->accxq && xq->accxq->panlist)
    {
      xq->accxq->panlist = g_list_remove (xq->accxq->panlist, xq);
    }

  if (GTK_IS_NOTEBOOK (xq->parent))
    {
      id = gtk_notebook_page_num (GTK_NOTEBOOK (xq->parent), xq->widget);
      gtk_notebook_remove_page (GTK_NOTEBOOK (xq->parent), id);
    }

  if (xq->qipu)
    {
      qipu_destroy (xq->qipu);
    }

  g_list_free (xq->cur_moves);
  xq->cur_moves = NULL;

  if (xq->parterthread)
    {
      g_thread_join (xq->parterthread);
    }
  session_delete (xq->session);
  if (xq->name)
    {
      g_free (xq->name);
    }
  free (xq);
}

void
accxq_delete (accxq_t * xq)
{
  GList *list;
  struct pan_node *pn;

  while ((list = g_list_first (xq->panlist)) != NULL)
    {
      pn = list->data;
      pannode_delete (pn);
      xq->panlist = g_list_remove (xq->panlist, list);
    }
  free (xq);
}

static gpointer
pn_black_move (gpointer arg)
{
  struct pan_node *xq;
  struct xq_move *move;

  xq = (struct pan_node *) arg;

  if (xq->state == PAN_STATE_GAME_OVER)
    {
      acc_debug (_("game is over\n"));
      return xq;
    }

  move = session_get_move (xq->session, xq->cur_fen, xq->cur_moves, g_set->pc_level, 0);
  if (move == NULL)
    {
      acc_err (_("Can't find a good move\n"));
      return xq;
    }

  gdk_threads_enter ();
  pn_move (xq, move->fromx, move->fromy, move->tox, move->toy);
  pn_refresh (xq);
  xq->state = PAN_STATE_TURN;
  gdk_threads_leave ();

  return xq;
}

static void 
pn_refresh (struct pan_node *pn)
{
  g_signal_emit_by_name (pn->drawarea, "expose-event", NULL, pn);
}

static void
pn_move (struct pan_node *xq, gint x, gint y, gint nx, gint ny)
{
  GtkWidget *dia;
  char mes[0x100];
  int isover;
  struct xq_move *move;
  GtkTreeIter iter[1];
  GtkTreePath *path;
  GtkTreeView *tree;

  acc_debug ("move [%s] [%d:%d] -> [%d:%d]\n", zi_chinese[(int) xq->cur_pan[y][x]], x, y, nx, ny);

  if (xq->move_count >= sizeof (xq->moves) / sizeof (xq->moves[0]))
    {
      acc_err ( _("move count error.\n"));
      return;
    }

  move = xq->moves + xq->move_count;
  xq->move_count ++;

  move->dead = xq->cur_pan[ny][nx];
  move->isred = IS_RED (xq->cur_pan[y][x]);
  move->fromx = x;
  move->fromy = y;
  move->tox = nx;
  move->toy = ny;

  move_to_desc (xq->cur_pan, move, mes, sizeof mes);
  gtk_list_store_append (GTK_LIST_STORE (xq->qipu_list), iter);
  gtk_list_store_set (GTK_LIST_STORE (xq->qipu_list), iter, 1, mes, -1);

  path = gtk_tree_model_get_path (GTK_TREE_MODEL (xq->qipu_list), iter);
  if (path)
    {
      tree = gtk_tree_selection_get_tree_view (GTK_TREE_SELECTION (xq->qipu_sel));
      gtk_tree_view_scroll_to_cell (tree, path, NULL, FALSE, 0, 0);
      gtk_tree_path_free (path);
    }

  pan_move (xq->cur_pan, x, y, nx, ny);

  if (move->dead)
    {
      xq->cur_moves = NULL;
      memcpy (xq->cur_fen, xq->cur_pan, 10 * 9 * sizeof (char));
      xq->cur_fen->half_count = 0;
      if (IS_RED (move->dead))
        {
          xq->cur_fen->is_redturn = 1;
        }
      else
        {
          xq->cur_fen->is_redturn = 0;
        }
    }
  else
    {
      xq->cur_moves = g_list_append (xq->cur_moves, move);
      xq->cur_fen->half_count ++;
    }

  isover = zgxq_is_over (xq->cur_pan);
  if (isover)
    {
      xq->state = PAN_STATE_GAME_OVER;

      g_snprintf (mes, sizeof mes, _("!!!Game over. You %s.\n"),
                  isover > 0 ? _("WIN") : _("LOSE"));
      dia =
        gtk_message_dialog_new (NULL,
                                GTK_DIALOG_DESTROY_WITH_PARENT,
                                GTK_MESSAGE_INFO, GTK_BUTTONS_CLOSE, mes);
      gtk_dialog_run (GTK_DIALOG (dia));
      gtk_widget_destroy (dia);
      g_signal_emit_by_name (xq->bstop, "clicked", xq);
    }
}

static void 
pn_unmove (struct pan_node *pn)
{
  struct xq_move * move;
  GtkTreeIter iter[1];
  gint nth;
  int id;
  GList *last;

  if (pn->move_count <= 0)
    {
      return;
    }

  pn->move_count --;
  move = pn->moves + pn->move_count;

  pn->cur_pan[move->fromy][move->fromx] = pn->cur_pan[move->toy][move->tox];
  pn->cur_pan[move->toy][move->tox] = move->dead;

  if (move->dead)
    {
      memcpy (pn->cur_fen->pan, pn->cur_pan, 10 * 9 * sizeof (char));
      pn->cur_fen->half_count = 0;
      pn->cur_fen->is_redturn = IS_BLACK (move->dead);
      pn->cur_moves = NULL;

      for (id = pn->move_count - 1, move = pn->moves + pn->move_count - 1;
           id >= 0 && move->dead == 0;
           id --, move = pn->moves + id)
        {
          pn->cur_fen->pan[move->fromy][move->fromx] = pn->cur_fen->pan[move->toy][move->tox];
          pn->cur_fen->pan[move->toy][move->tox] = move->dead;

          pn->cur_moves = g_list_prepend (pn->cur_moves, move);
          pn->cur_fen->is_redturn = IS_RED (pn->cur_fen->pan[move->fromy][move->fromx]);
          pn->cur_fen->half_count ++;
        }
    }
  else
    {
      pn->cur_fen->half_count --;

      last = g_list_last (pn->cur_moves);
      pn->cur_moves = g_list_remove_link (pn->cur_moves, last);
    }

  nth = gtk_tree_model_iter_n_children (GTK_TREE_MODEL (pn->qipu_list), NULL);
  gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (pn->qipu_list), iter, NULL, nth - 1);
  gtk_list_store_remove (GTK_LIST_STORE (pn->qipu_list), iter);
}

static void
pn_redraw_cb (GtkWidget * wid, GdkEventExpose * pose, struct pan_node *xq)
{
  gint usew, useh, isw, sub, i, j, midy;
  GdkGC *gc;
  GdkColor color[1];

  if (!GDK_IS_DRAWABLE (wid->window))
    {
      return;
    }

//  acc_debug ("drawarea allocation '%d:%d'\n", wid->allocation.width, wid->allocation.height);
  if (wid->allocation.width < wid->allocation.height)
    {
      isw = 1;
      usew = wid->allocation.width;
      useh = wid->allocation.width;
      sub = wid->allocation.height - wid->allocation.width;
    }
  else
    {
      isw = 0;
      usew = wid->allocation.height;
      useh = wid->allocation.height;
      sub = wid->allocation.width - wid->allocation.height;
    }

  xq->perx = usew / 9;
  xq->pery = useh / 11;
  if (isw)
    {
      xq->fromx = xq->perx >> 1;
      xq->fromy = (sub >> 1) + xq->pery;
    }
  else
    {
      xq->fromx = (sub >> 1) + (xq->perx >> 1);
      xq->fromy = xq->pery;
    }

  xq->tox = xq->fromx + xq->perx * 8;
  xq->toy = xq->fromy + xq->pery * 9;

  midy = xq->fromy + 4 * xq->pery;
  xq->qix = xq->perx - (xq->perx >> 2);
  xq->qiy = xq->perx - (xq->pery >> 2);

  gc = gdk_gc_new (wid->window);

  /*
   * Draw the background
   * */
  color->red = 0;
  color->green = 0;
  color->blue = 0;
  gdk_gc_set_rgb_bg_color (gc, color);
  gdk_gc_set_rgb_fg_color (gc, color);
  gdk_draw_rectangle (wid->window, gc,
                      TRUE,
                      0, 0, wid->allocation.width, wid->allocation.height);

  /*
   * Draw the 10 * 9 lines
   * */
  color->red = 32768;
  color->green = 32768;
  color->blue = 32768;
  gdk_gc_set_rgb_bg_color (gc, color);
  gdk_gc_set_rgb_fg_color (gc, color);

  gdk_gc_set_line_attributes (gc, 3, 0, 0, 0);
  for (i = 0; i < 10; ++i)
    {
      gdk_draw_line (wid->window, gc, xq->fromx, xq->fromy + i * xq->pery,
                     xq->tox, xq->fromy + i * xq->pery);
    }

  for (j = 1; j < 8; ++j)
    {
      gdk_draw_line (wid->window, gc, xq->fromx + j * xq->perx, xq->fromy,
                     xq->fromx + j * xq->perx, midy);
      gdk_draw_line (wid->window, gc, xq->fromx + j * xq->perx,
                     midy + xq->pery, xq->fromx + j * xq->perx, xq->toy);
    }

  gdk_draw_line (wid->window, gc, xq->fromx, xq->fromy, xq->fromx, xq->toy);
  gdk_draw_line (wid->window, gc, xq->fromx + 8 * xq->perx, xq->fromy,
                 xq->fromx + 8 * xq->perx, xq->toy);

  /* Draw small text */
  for (j = 0; j < 9; ++ j)
    {
      pn_draw_text_at (xq, xq->fromx + j * xq->perx - 3, xq->fromy - (xq->qiy >> 1) - 18, number_str[j + 1]);
    }

  /* Draw big text */
  for (j = 0; j < 9; ++ j)
    {
      pn_draw_text_at (xq, xq->fromx + j * xq->perx - 6, xq->toy + (xq->qiy >> 1) + 3, number_chinese_str[9 - j]);
    }

  /*
   * Draw the X in home
   * */
  gdk_gc_set_line_attributes (gc, 2, 0, 0, 0);
  gdk_draw_line (wid->window, gc, xq->fromx + 3 * xq->perx, xq->fromy,
                 xq->fromx + 5 * xq->perx, xq->fromy + 2 * xq->pery);
  gdk_draw_line (wid->window, gc, xq->fromx + 5 * xq->perx, xq->fromy,
                 xq->fromx + 3 * xq->perx, xq->fromy + 2 * xq->pery);
  gdk_draw_line (wid->window, gc, xq->fromx + 3 * xq->perx,
                 xq->toy - xq->pery * 2, xq->fromx + 5 * xq->perx, xq->toy);
  gdk_draw_line (wid->window, gc, xq->fromx + 5 * xq->perx,
                 xq->toy - xq->pery * 2, xq->fromx + 3 * xq->perx, xq->toy);

  g_object_unref (gc);

  for (i = 0; i < 10; ++i)
    {
      for (j = 0; j < 9; ++j)
        {
          if (xq->cur_pan[i][j])
            {
              pn_draw_zi_at (xq,
                             xq->fromx + xq->perx * (j) - (xq->qix >> 1),
                             xq->fromy + xq->pery * (i) - (xq->qiy >> 1),
                             xq->cur_pan[i][j]);
            }
        }
    }
}

static gboolean
pn_press_cb (GtkWidget * wid, GdkEventButton * ev, struct pan_node *xq)
{
  gint x, y;

  if (xq->state != PAN_STATE_TURN
      && xq->state != PAN_STATE_INIT)
    {
      return FALSE;
    }

  if (ev->button != 1)
    {
      return FALSE;
    }

  x = (gint) (ev->x - xq->fromx + (xq->perx >> 1)) / xq->perx;
  y = (gint) (ev->y - xq->fromy + (xq->pery >> 1)) / xq->pery;
  if (xq->cur_pan[y][x] == 0)
    {
      return FALSE;
    }

  if (xq->state != PAN_STATE_INIT
      && xq->other_type == OTHER_PC 
      && IS_BLACK (xq->cur_pan[y][x]))
    {
      return FALSE;
    }

  acc_debug ("you selected [%d]\n", xq->cur_pan[y][x]);
  if (xq->state == PAN_STATE_TURN)
    {
      xq->state = PAN_STATE_TURN_SELECTED;
    }
  xq->selx = x;
  xq->sely = y;
  xq->motionx = xq->fromx + x * xq->perx - (xq->qix >> 1);
  xq->motiony = xq->fromy + y * xq->pery - (xq->qiy >> 1);

  return TRUE;
}

static gboolean
pn_release_cb (GtkWidget * wid, GdkEventButton * ev, struct pan_node *xq)
{
  gint x, y;

  if (xq->state != PAN_STATE_TURN_SELECTED
      && xq->state != PAN_STATE_INIT)
    {
      return FALSE;
    }

  if (ev->button != 1)
    {
      return FALSE;
    }

  x = (gint) (ev->x - xq->fromx + (xq->perx >> 1)) / xq->perx;
  y = (gint) (ev->y - xq->fromy + (xq->pery >> 1)) / xq->pery;

  if (xq->state == PAN_STATE_INIT)
    {
      if (x == xq->selx
          && y == xq->sely)
        {
          xq->motionx = xq->motiony = 0;
          return FALSE;
        }
      pan_move (xq->cur_pan, xq->selx, xq->sely, x, y);
      pn_refresh (xq);
    }

  else if (zgxq_move_is_valid (xq->cur_pan, xq->selx, xq->sely, x, y))
    {
      struct xq_move move[1];

      move->fromx = xq->selx;
      move->fromy = xq->sely;
      move->tox = x;
      move->toy = y;

      pn_move (xq, xq->selx, xq->sely, x, y);
      pn_refresh (xq);

      if (xq->state == PAN_STATE_GAME_OVER)
        {
          xq->motionx = xq->motiony = 0;
          return TRUE;
        }

      if (xq->other_type == OTHER_PC)
        {
          xq->state = PAN_STATE_NOT_TURN;

          if (xq->parterthread != NULL)
            {
              g_thread_join (xq->parterthread);
            }
          xq->parterthread = g_thread_create (pn_black_move, xq, TRUE, NULL);
        }
      else
        {
          xq->state = PAN_STATE_TURN;
        }
    }
  else
    {
      pn_refresh (xq);
      xq->state = PAN_STATE_TURN;
    }

  xq->motionx = xq->motiony = 0;

  return TRUE;
}

static void 
pan_baiqi_cb (GtkWidget * wid, struct pan_node *pn)
{
  g_list_free (pn->cur_moves);
  pn->cur_moves = NULL;

  memcpy (pn->cur_pan, default_pan, 10 * 9 * sizeof (char));

  if (GTK_IS_LIST_STORE (pn->qipu_list))
    {
      gtk_list_store_clear (GTK_LIST_STORE (pn->qipu_list));
    }
  qipu_destroy (pn->qipu);
  pn->qipu = qipu_new ();

  gtk_widget_set_sensitive (wid, FALSE);
  gtk_widget_set_sensitive (pn->bstart, TRUE);
  gtk_widget_set_sensitive (pn->bhui, FALSE);
  gtk_widget_set_sensitive (pn->bstop, FALSE);

  pn->state = PAN_STATE_INIT;

  pn_refresh (pn);
}

static void
pan_start_cb (GtkWidget * wid, struct pan_node *pn)
{
  pn->state = PAN_STATE_TURN;

  memcpy (pn->fen, pn->cur_pan, 10 * 9 * sizeof (char));
  memcpy (pn->cur_fen, pn->cur_pan, 10 * 9 * sizeof (char));
  pn->cur_fen->half_count = 0;
  pn->cur_fen->is_redturn = 1;
  pn->cur_moves = NULL;

  gtk_widget_set_sensitive (wid, FALSE);
  gtk_widget_set_sensitive (pn->bbaiqi, TRUE);
  gtk_widget_set_sensitive (pn->bhui, TRUE);
  gtk_widget_set_sensitive (pn->bstop, TRUE);
}

static void
pan_hui_cb (GtkWidget * wid, struct pan_node *xq)
{
  if (xq->state != PAN_STATE_TURN)
    {
      acc_debug (_("Can't hui qi now\n"));
      return;
    }

  if (xq->move_count <= 1)
    {
      return;
    }

  pn_unmove (xq);

  pn_unmove (xq);

  pn_refresh (xq);
}

static void
pan_stop_cb (GtkWidget * wid, struct pan_node *xq)
{
  xq->state = PAN_STATE_GAME_OVER;
  gtk_widget_set_sensitive (xq->bbaiqi, TRUE);
  gtk_widget_set_sensitive (xq->bstart, TRUE);
  gtk_widget_set_sensitive (xq->bhui, FALSE);
  gtk_widget_set_sensitive (wid, FALSE);
}

static void
pan_quit_cb (GtkWidget * wid, struct pan_node *xq)
{
  pannode_delete (xq);
}

static GdkPixbuf *
xq_qi_buf (int id, gint x, gint y)
{
  GError *error = NULL;
  char path[PATH_MAX];
  static const char *filenames[] =
  {
    "",
    "rshuai.bmp",
    "bjiang.bmp",
    "rshi.bmp",
    "bshi.bmp",
    "rxiang.bmp",
    "bxiang.bmp",
    "rma.bmp",
    "bma.bmp",
    "rju.bmp",
    "bju.bmp",
    "rpao.bmp",
    "bpao.bmp",
    "rbing.bmp",
    "bzu.bmp",
  };

  static GdkPixbuf *pix[15] = { 0 };

  if (id <= 0 || id > B_BING || x < 10 || y < 10)
    {
      return NULL;
    }

  if (pix[id] == NULL)
    {
      g_snprintf (path, sizeof path, "%s/%s/%s", ACNCHESS_DATADIR, "images", filenames[(int) id]);
      pix[id] = gdk_pixbuf_new_from_file (path, &error);

      if (error)
        {
          acc_debug ("Load error: %s\n", error->message);
          g_error_free (error);
        }
    }

  if (pix[id])
    {
      GdkPixbuf *newpix;

      newpix = gdk_pixbuf_scale_simple (pix[id], x, y, GDK_INTERP_BILINEAR);
      if (newpix)
        {
          return newpix;
        }
      else
        {
          return gdk_pixbuf_copy (pix[id]);
        }
    }

  return NULL;
}

static int
xq_append_pan (accxq_t * xq, struct pan_node *pn)
{
  int id;
  char buf[0x200];
  GtkWidget *hbox, *label, *button;

  xq->panlist = g_list_append (xq->panlist, pn);

  hbox = gtk_hbox_new (FALSE, 0);
  id = gtk_notebook_get_n_pages (GTK_NOTEBOOK (xq->pantab)) + 1;
  g_snprintf (buf, sizeof buf, "%s:%d", pn->name, id);
  label = gtk_label_new (buf);
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
  button = gtk_button_new_with_label ("x");
  g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (pan_quit_cb),
                    pn);
  gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);
  gtk_widget_show_all (hbox);

  pn->parent = xq->pantab;
  id = gtk_notebook_append_page (GTK_NOTEBOOK (xq->pantab), pn->widget, hbox);

  return id;
}

static void
accxq_onload (accxq_t * xq)
{
  session_t *ses;

  ses = session_new (g_set->pc_engine);
  if (ses == NULL)
    {
      acc_err (_("Memory error\n"));
      return;
    }

  session_start (ses);

  pannode_new (xq, ses);
}

static void
pn_draw_zi_at (struct pan_node *xq, gint zx, gint zy, gint zi)
{
  GdkPixbuf *pix;

  pix = xq_qi_buf (zi, xq->qix, xq->qiy);
  if (pix != NULL)
    {
      gdk_draw_pixbuf (xq->drawarea->window,
                       NULL,
                       pix,
                       0, 0,
                       zx,
                       zy,
                       -1, -1,
                       GDK_RGB_DITHER_NORMAL, 0, 0);
      g_object_unref (pix);
    }
}

static void
pn_draw_text_at (struct pan_node *pn, gint x, gint y, const gchar *text)
{
  GdkGC *gc;
  GdkColor color[1];
  PangoLayout *layout;
  gchar *utext;

  gc = gdk_gc_new (pn->drawarea->window);

  color->red = 65535;
  color->green = 65535;
  color->blue = 65535;
  gdk_gc_set_rgb_bg_color (gc, color);
  gdk_gc_set_rgb_fg_color (gc, color);
  gdk_gc_set_line_attributes (gc, 3, 0, 0, 0);

  utext = g_locale_to_utf8 (text, -1, NULL, NULL, NULL);
  if (utext)
    {
      layout = gtk_widget_create_pango_layout (pn->drawarea, utext);
      g_free (utext);
    }
  else
    {
      layout = gtk_widget_create_pango_layout (pn->drawarea, text);
    }

  if (layout)
    {
      gdk_draw_layout (pn->drawarea->window,
                       gc,
                       x,
                       y,
                       layout);
      g_object_unref (layout);
    }

  g_object_unref (gc);
}

static void
pn_other_changed (GtkWidget * wid, struct pan_node *pn)
{
  gint id;

  id = gtk_combo_box_get_active (GTK_COMBO_BOX (wid));
  pn->other_type = id;
}

static GtkWidget * 
pan_other_new (struct pan_node *pn)
{
  GtkWidget *vbox, *label;

  vbox = gtk_vbox_new (FALSE, 1);

  label = gtk_label_new (_ ("enemy"));
  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 1);

  return vbox;
}

static GtkWidget * 
pan_myself_new (struct pan_node *pn)
{
  GtkWidget *vbox, *label;

  vbox = gtk_vbox_new (FALSE, 1);

  label = gtk_label_new (_ ("me"));
  gtk_box_pack_end (GTK_BOX (vbox), label, FALSE, FALSE, 1);

  return vbox;
}

static GtkWidget *
pan_draw_ui (accxq_t *xq, struct pan_node *pn)
{
  GtkWidget *vbox, *hbox, *other, *myself;

  vbox = gtk_vbox_new (FALSE, 2);

  hbox = gtk_hbox_new (FALSE, 1);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 5);

  other = pan_other_new (pn);
  myself = pan_myself_new (pn);
  gtk_box_pack_start (GTK_BOX (hbox), other, FALSE, FALSE, 1);
  gtk_box_pack_end (GTK_BOX (hbox), myself, FALSE, FALSE, 1);

  pn->drawarea = gtk_drawing_area_new ();
  gtk_box_pack_start (GTK_BOX (hbox), pn->drawarea, TRUE, TRUE, 5);

  gtk_widget_add_events (pn->drawarea,
                         GDK_ALL_EVENTS_MASK);
  g_signal_connect_after (G_OBJECT (pn->drawarea), "expose-event",
                          G_CALLBACK (pn_redraw_cb), pn);
  g_signal_connect (G_OBJECT (pn->drawarea), "button-press-event",
                    G_CALLBACK (pn_press_cb), pn);
  g_signal_connect (G_OBJECT (pn->drawarea), "button-release-event",
                    G_CALLBACK (pn_release_cb), pn);

  hbox = gtk_hbox_new (TRUE, 5);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 5);

  pn->bbaiqi = gtk_button_new_with_label (_("Baiqi"));
  gtk_box_pack_start (GTK_BOX (hbox), pn->bbaiqi, FALSE, FALSE, 5);
  g_signal_connect (G_OBJECT (pn->bbaiqi), "clicked",
                    G_CALLBACK (pan_baiqi_cb), pn);

  pn->bstart = gtk_button_new_with_label (_("Start"));
  gtk_box_pack_start (GTK_BOX (hbox), pn->bstart, FALSE, FALSE, 5);
  g_signal_connect (G_OBJECT (pn->bstart), "clicked",
                    G_CALLBACK (pan_start_cb), pn);

  pn->bhui = gtk_button_new_with_label (_("Hui"));
  gtk_box_pack_start (GTK_BOX (hbox), pn->bhui, FALSE, FALSE, 5);
  g_signal_connect (G_OBJECT (pn->bhui), "clicked", G_CALLBACK (pan_hui_cb),
                    pn);

  pn->bstop = gtk_button_new_with_label (_("Stop"));
  gtk_box_pack_start (GTK_BOX (hbox), pn->bstop, FALSE, FALSE, 5);
  g_signal_connect (G_OBJECT (pn->bstop), "clicked", G_CALLBACK (pan_stop_cb),
                    pn);

  g_signal_emit_by_name (pn->bbaiqi, "clicked", xq);

  return vbox;
}

static void
pn_control_changed (GtkWidget * wid, struct pan_node *pn)
{
  gint id;

  id = gtk_combo_box_get_active (GTK_COMBO_BOX (wid));
  if (id != pn->conf_id)
    {
      gtk_container_remove (GTK_CONTAINER (pn->wconf), pn->conf[pn->conf_id]);
      pn->conf_id = id;
      gtk_container_add (GTK_CONTAINER (pn->wconf), pn->conf[id]);
      gtk_widget_show_all (pn->conf[id]);
    }
}

static GtkWidget *
pan_control_ui (accxq_t *xq, struct pan_node *pn)
{
  GtkWidget *vbox, *hbox, *label, *combo;

  vbox = gtk_vbox_new (FALSE, 4);

  hbox = gtk_hbox_new (FALSE, 4);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
  label = gtk_label_new (_("Mode"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 4);
  combo = gtk_combo_box_new_text ();
  gtk_box_pack_start (GTK_BOX (hbox), combo, TRUE, TRUE, 4);

  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("Common"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("Qipu"));

  pn->wconf = gtk_frame_new ("");
  gtk_container_add (GTK_CONTAINER (vbox), pn->wconf);

  pn->conf[0] = pn_pcset_new (pn);
  g_object_ref (pn->conf[0]);

  pn->conf[1] = pn_qpset_new (pn);
  g_object_ref (pn->conf[1]);

  gtk_container_add (GTK_CONTAINER (pn->wconf), pn->conf[0]);
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);
  g_signal_connect (G_OBJECT (combo), "changed", G_CALLBACK (pn_control_changed), pn);

  return vbox;
}

static gboolean
pn_pcset_level_changed (GtkRange *range,
                        struct pan_node *pn)
{
  int level;

  level = (int) gtk_range_get_value (range);
  g_set->pc_level = level;

  return TRUE;
}

static char *
pn_pcset_level_format (GtkScale *scale,
                       gdouble value)
{
  return g_strdup_printf (_ ("level %d"), (int) value);
}

static GtkWidget *
pn_pcset_new (struct pan_node *pn)
{
  GtkWidget *vbox, *hbox, *combo, *label, *eng, *engexp, *scale;

  vbox = gtk_vbox_new (FALSE, 4);

  combo = gtk_combo_box_new_text ();
  gtk_box_pack_start (GTK_BOX (vbox), combo, FALSE, FALSE, 1);
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _ ("Bot"));
  gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _ ("Manual"));
  gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);
  g_signal_connect (G_OBJECT (combo), "changed", G_CALLBACK (pn_other_changed), pn);

  hbox = gtk_hbox_new (FALSE, 4);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new (_ ("Engine"));
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2);
  eng = gtk_entry_new ();
  gtk_entry_set_text (GTK_ENTRY (eng), g_set->pc_engine);
  gtk_box_pack_start (GTK_BOX (hbox), eng, TRUE, TRUE, 2);
  engexp = gtk_button_new_with_label (_ ("Explore"));
  gtk_box_pack_end (GTK_BOX (hbox), engexp, FALSE, FALSE, 2);
  g_object_set_data (G_OBJECT (engexp), "entry", eng);
  g_signal_connect (G_OBJECT (engexp), "clicked", G_CALLBACK (pn_pcset_engine_select), pn);

  hbox = gtk_hbox_new (FALSE, 4);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 2);

  scale = gtk_hscale_new_with_range (1, 7, 1);
  gtk_range_set_value (GTK_RANGE (scale), g_set->pc_level);
  gtk_scale_set_digits (GTK_SCALE (scale), 0);
  gtk_box_pack_end (GTK_BOX (hbox), scale, TRUE, TRUE, 4);

  g_signal_connect (G_OBJECT (scale), "value-changed", G_CALLBACK (pn_pcset_level_changed), pn);
  g_signal_connect (G_OBJECT (scale), "format-value", G_CALLBACK (pn_pcset_level_format), pn);

  return vbox;
}

static GtkWidget *
pn_qpset_new (struct pan_node * pn)
{
  GtkWidget *vbox, *vbbox, *win, *tree; 
  GtkWidget *bload, *bsave, *bstart, *bbranch, *bnext;
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  vbox = gtk_vbox_new (FALSE, 4);

  win = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (win),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (win),
                                       GTK_SHADOW_IN);
  gtk_box_pack_start (GTK_BOX (vbox), win, TRUE, TRUE, 1);

  pn->qipu_list = gtk_list_store_new (2, G_TYPE_POINTER, G_TYPE_STRING);
  tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (pn->qipu_list));
  gtk_container_add (GTK_CONTAINER (win), tree);
  g_signal_connect_after (G_OBJECT (tree), "row-activated", G_CALLBACK (qp_row_activated_cb), pn);

  pn->qipu_sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
  gtk_tree_selection_set_mode (pn->qipu_sel, GTK_SELECTION_SINGLE);

  /* Nick Name Column */
  renderer = gtk_cell_renderer_text_new ();
  column =
    gtk_tree_view_column_new_with_attributes (_("step"), renderer,
        "text", 1, NULL);
  gtk_tree_view_column_set_resizable (column, TRUE);
  gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);

  vbbox = gtk_vbox_new (TRUE, 2);
  gtk_box_pack_start (GTK_BOX (vbox), vbbox, FALSE, FALSE, 1);

  bload = gtk_button_new_with_label (_ ("load"));
  gtk_box_pack_start (GTK_BOX (vbbox), bload, FALSE, FALSE, 2);
  g_signal_connect (G_OBJECT (bload), "clicked", G_CALLBACK (qp_load_cb), pn);

  bsave = gtk_button_new_with_label (_ ("save"));
  gtk_box_pack_start (GTK_BOX (vbbox), bsave, FALSE, FALSE, 2);
  g_signal_connect (G_OBJECT (bsave), "clicked", G_CALLBACK (qp_save_cb), pn);

  bstart = gtk_button_new_with_label (_ ("begin"));
  gtk_box_pack_start (GTK_BOX (vbbox), bstart, FALSE, FALSE, 2);
  g_signal_connect (G_OBJECT (bstart), "clicked", G_CALLBACK (qp_start_cb), pn);

  bbranch = gtk_button_new_with_label (_ ("branch"));
  gtk_box_pack_start (GTK_BOX (vbox), bbranch, FALSE, FALSE, 2);
  g_signal_connect (G_OBJECT (bbranch), "clicked", G_CALLBACK (qp_branch_cb), pn);

  bnext = gtk_button_new_with_label (_ ("next"));
  gtk_box_pack_start (GTK_BOX (vbbox), bnext, FALSE, FALSE, 2);
  g_signal_connect (G_OBJECT (bnext), "clicked", G_CALLBACK (qp_next_cb), pn);

  return vbox;
}

static void
qp_load_cb (GtkWidget *wid, struct pan_node *pn)
{
  GtkWidget *dialog;

  dialog = gtk_file_chooser_dialog_new ("Choose File", NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                        NULL);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      gchar *filename;

      filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
      if (filename)
        {
          if (pn->qipu)
            {
              qipu_destroy (pn->qipu);
              pn->qipu = NULL;
            }

          gtk_list_store_clear (GTK_LIST_STORE (pn->qipu_list));

          pn->qipu = qipu_new_from_file (filename);
          if (pn->qipu)
            {
              int ret, isred;
              struct xq_move move[1];
              char mes[0x100], pan[10][9];
              GtkTreeIter iter[1];

              qipu_get_fen (pn->qipu, pn->fen);
              memcpy (pan, pn->fen, 10 * 9 * sizeof (char));

              for (isred = 1, ret = qipu_move_first (pn->qipu, move);
                   ret == 0;
                   ret = qipu_move_next (pn->qipu, NULL, move))
                {
                  move->isred = isred;
                  isred = isred ? 0: 1;
                  move_to_desc (pan, move, mes, sizeof mes);
                  pan_move (pan, move->fromx, move->fromy, move->tox, move->toy);
                  gtk_list_store_append (GTK_LIST_STORE (pn->qipu_list), iter);
                  gtk_list_store_set (GTK_LIST_STORE (pn->qipu_list), iter, 1, mes, -1);
                }

              pn_refresh (pn);

              gtk_widget_set_sensitive (pn->bstart, FALSE);
              gtk_widget_set_sensitive (pn->bbaiqi, TRUE);
              gtk_widget_set_sensitive (pn->bhui, FALSE);

              pn->state = PAN_STATE_QIPU;
            }
          g_free (filename);
        }
    }

  gtk_widget_destroy (dialog);
}

static void
qp_save_dialog_file_set (GtkFileChooserButton *but, GtkWidget *entry)
{
  GtkWidget *dialog;

  dialog = gtk_file_chooser_dialog_new (_ ("Save as"), NULL, GTK_FILE_CHOOSER_ACTION_SAVE,
                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                        NULL);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      const gchar *filename;
      filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
      gtk_entry_set_text (GTK_ENTRY (entry), filename);
    }
  gtk_widget_destroy (dialog);
}

static void
qp_save_cb (GtkWidget *wid, struct pan_node *pn)
{
  GtkWidget *dialog;
  GtkWidget *hbox, *glabel, *gevent, *gdate, *gsite, *gcom, *gred, *gblack, *gname, *gexplore;

  dialog = gtk_dialog_new_with_buttons (_ ("Save qipu"), NULL,
                                        GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                                        GTK_STOCK_OK,
                                        GTK_RESPONSE_ACCEPT,
                                        GTK_STOCK_CANCEL,
                                        GTK_RESPONSE_REJECT,
                                        NULL);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, FALSE, FALSE, 5);
  glabel = gtk_label_new (_ ("Event"));
  gtk_widget_set_size_request (GTK_WIDGET (glabel), 120, 30);
  gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 5);
  gevent = gtk_entry_new ();
  gtk_box_pack_end (GTK_BOX (hbox), gevent, TRUE, TRUE, 5);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, FALSE, FALSE, 5);
  glabel = gtk_label_new (_ ("Date"));
  gtk_widget_set_size_request (GTK_WIDGET (glabel), 120, 30);
  gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 5);
  gdate = gtk_entry_new ();
  gtk_box_pack_end (GTK_BOX (hbox), gdate, TRUE, TRUE, 5);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, FALSE, FALSE, 5);
  glabel = gtk_label_new (_ ("Site"));
  gtk_widget_set_size_request (GTK_WIDGET (glabel), 120, 30);
  gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 5);
  gsite = gtk_entry_new ();
  gtk_box_pack_end (GTK_BOX (hbox), gsite, TRUE, TRUE, 5);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, FALSE, FALSE, 5);
  glabel = gtk_label_new (_ ("Red"));
  gtk_widget_set_size_request (GTK_WIDGET (glabel), 120, 30);
  gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 5);
  gred = gtk_entry_new ();
  gtk_box_pack_end (GTK_BOX (hbox), gred, TRUE, TRUE, 5);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, FALSE, FALSE, 5);
  glabel = gtk_label_new (_ ("Black"));
  gtk_widget_set_size_request (GTK_WIDGET (glabel), 120, 30);
  gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 5);
  gblack = gtk_entry_new ();
  gtk_box_pack_end (GTK_BOX (hbox), gblack, TRUE, TRUE, 5);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 5);
  glabel = gtk_label_new (_ ("Commet"));
  gtk_widget_set_size_request (GTK_WIDGET (glabel), 120, 30);
  gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 5);
  gcom = gtk_text_view_new ();
  gtk_box_pack_end (GTK_BOX (hbox), gcom, TRUE, TRUE, 5);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, FALSE, FALSE, 5);
  glabel = gtk_label_new (_ ("FileName"));
  gtk_widget_set_size_request (GTK_WIDGET (glabel), 120, 30);
  gtk_box_pack_start (GTK_BOX (hbox), glabel, FALSE, FALSE, 5);
  gname = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox), gname, TRUE, TRUE, 2);
  gexplore = gtk_button_new_with_label (_ ("Explore ..."));
  g_signal_connect (G_OBJECT (gexplore), "clicked", G_CALLBACK (qp_save_dialog_file_set), gname);
  gtk_box_pack_end (GTK_BOX (hbox), gexplore, FALSE, FALSE, 2);

  gtk_widget_show_all (GTK_DIALOG (dialog)->vbox);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      const gchar *text; 
      gchar *commet;
      GtkTextBuffer *gbuffer;
      GtkTextIter siter[1], eiter[1];
      gboolean has;
      GtkTreeIter iter[1];
      char pan[10][9];
      struct xq_move move[1];

      text = gtk_entry_get_text (GTK_ENTRY (gevent));
      qipu_set_info (pn->qipu, "Event", text);
      text = gtk_entry_get_text (GTK_ENTRY (gdate));
      qipu_set_info (pn->qipu, "Date", text);
      text = gtk_entry_get_text (GTK_ENTRY (gsite));
      qipu_set_info (pn->qipu, "Site", text);
      text = gtk_entry_get_text (GTK_ENTRY (gred));
      qipu_set_info (pn->qipu, "Red", text);
      text = gtk_entry_get_text (GTK_ENTRY (gblack));
      qipu_set_info (pn->qipu, "Black", text);

      gbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (gcom));
      gtk_text_buffer_get_start_iter (gbuffer, siter);
      gtk_text_buffer_get_end_iter (gbuffer, eiter);
      commet = gtk_text_buffer_get_text (gbuffer, siter, eiter, TRUE);
      qipu_set_info (pn->qipu, "Commet", commet);
      g_free (commet);

      memcpy (pan, pn->fen->pan, 10 * 9 * sizeof (char));
      move->isred = 1;
      for (has = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (pn->qipu_list), iter);
           has == TRUE;
           has = gtk_tree_model_iter_next (GTK_TREE_MODEL (pn->qipu_list), iter))
        {
          gtk_tree_model_get (GTK_TREE_MODEL (pn->qipu_list), iter, 1, &text, -1);
          desc_to_move (pan, text, move);
          qipu_append_move (pn->qipu, move);
          pan_move (pan, move->fromx, move->fromy, move->tox, move->toy);
          move->isred = (move->isred ? 0: 1);
        }

      text = gtk_entry_get_text (GTK_ENTRY (gname));
      qipu_save (pn->qipu, text);
    }

  gtk_widget_destroy (dialog);
}

static void
qp_start_cb (GtkWidget *wid, struct pan_node *pn)
{
  int has;
  struct xq_move move[1];
  GtkTreeIter iter[1];

  qipu_set_fen (pn->qipu, pn->fen);

  has = qipu_move_first (pn->qipu, move);
  if (has < 0)
    {
      acc_debug ("No first step\n");
      return;
    }

  pan_move (pn->cur_pan, move->fromx, move->fromy, move->tox, move->toy);
  pn_refresh (pn);

  gtk_tree_model_get_iter_first (GTK_TREE_MODEL (pn->qipu_list), iter);
  gtk_tree_selection_select_iter (GTK_TREE_SELECTION (pn->qipu_sel), iter);
}

static void 
qp_branch_cb (GtkWidget *wid, struct pan_node *pn)
{
}

static void
qp_next_cb (GtkWidget *wid, struct pan_node *pn)
{
  int has;
  struct xq_move move[1];
  GtkTreeIter iter[1];
  GtkTreeView *tree;
  GtkTreePath *path;

  has = qipu_move_next (pn->qipu, NULL, move);
  if (has < 0)
    {
      acc_debug ("No next step\n");
      return;
    }

  pan_move (pn->cur_pan, move->fromx, move->fromy, move->tox, move->toy);
  pn_refresh (pn);

  gtk_tree_selection_get_selected (GTK_TREE_SELECTION (pn->qipu_sel), NULL, iter);
  gtk_tree_model_iter_next (GTK_TREE_MODEL (pn->qipu_list), iter);
  gtk_tree_selection_select_iter (GTK_TREE_SELECTION (pn->qipu_sel), iter);

  path = gtk_tree_model_get_path (GTK_TREE_MODEL (pn->qipu_list), iter);
  if (path)
    {
      tree = gtk_tree_selection_get_tree_view (GTK_TREE_SELECTION (pn->qipu_sel));
      gtk_tree_view_scroll_to_cell (tree, path, NULL, FALSE, 0, 0);
      gtk_tree_path_free (path);
    }
}

static void 
qp_row_activated_cb (GtkTreeView *tree, GtkTreePath *select_path, GtkTreeViewColumn *col, struct pan_node *pn)
{
  struct xq_move move[1];
  GtkTreeIter iter[1];
  GtkTreePath *path;

  gtk_tree_model_get_iter_first (GTK_TREE_MODEL (pn->qipu_list), iter);
  path = gtk_tree_model_get_path (GTK_TREE_MODEL (pn->qipu_list), iter);
  if (path == NULL)
    {
      acc_err (_ ("Memory error.\n"));
      return;
    }

  qipu_get_fen (pn->qipu, pn->fen);
  qipu_move_first (pn->qipu, move);
  pan_move (pn->cur_pan, move->fromx, move->fromy, move->tox, move->toy);

  while (gtk_tree_path_compare (path, select_path) < 0)
    {
      gtk_tree_path_next (path);
      qipu_move_next (pn->qipu, NULL, move);
      pan_move (pn->cur_pan, move->fromx, move->fromy, move->tox, move->toy);
    }

  pn_refresh (pn);
}

static void 
pn_pcset_engine_select (GtkWidget *wid, struct pan_node *pn)
{
  GtkWidget *dialog;
  GtkWidget *entry;
  gchar *filename;

  dialog = gtk_file_chooser_dialog_new ("Choose File", NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                        NULL);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_ACCEPT)
    {
      return;
    }

  filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
  if (filename == NULL)
    {
      return;
    }

  gtk_widget_destroy (dialog);

  if (pn->session)
    {
      session_delete (pn->session);
    }

  entry = g_object_get_data (G_OBJECT (wid), "entry");
  gtk_entry_set_text (GTK_ENTRY (entry), filename);
  pn->session = session_new (filename);

  g_free (filename);

  g_set->pc_engine = (char *) gtk_entry_get_text (GTK_ENTRY (entry));
}
