/* * This file is part of the acnchess package
 * Copyright (C) <2010>  <Alf>
 *
 * Contact: Alf <naihe2010@126.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
/* @CFILE acnguy.c
 *
 *  Author: Alf <naihe2010@126.com>
 */
/* @date Created: 2011/07/13 10:08:48 Alf*/

#include "acnkern.h"
#include "ucci.h"
#include "iccs.h"
#include "fen.h"
#include "zgxq.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <aclapi.h>
#else
#include <pthread.h>
#endif

enum
{
  ACNGUY_FD_READ = 0,
  ACNGUY_FD_WRITE,
  ACNGUY_FD_ERROR,
};

enum acnguy_state
{
  ACNGUY_INIT,
  ACNGUY_WAIT,
  ACNGUY_GOING,
};

struct acnguy
{
  FILE *read_fp; 
  FILE *write_fp; 
  FILE *error_fp;

  int state;

  acnkern_t *search;

  fen_t fen[1];
};

static int acnguy_process (struct acnguy *, char *);
static int acnguy_process_input (struct acnguy *);
static int acnguy_engine_callback (struct acnguy *);

#define ACNGUY_INBUF_SIZE       0x2000  /* 8K maybe enough */

int
main (int argc, char *argv[])
{
  struct acnguy acnguy[1];
#ifdef WIN32
  HANDLE handle;
  DWORD pth, ret;
#else
  pthread_t pth;
#endif

  acnguy->read_fp = stdin;
  acnguy->write_fp = stdout;
  acnguy->error_fp = stderr;

  acnguy->search = acnkern_new ((acnkern_callback) acnguy_engine_callback, acnguy);
  if (acnguy->search == NULL)
    {
      fprintf (acnguy->error_fp, "The fd of acnguy maybe have some error.\n");
      exit (1);
    }

  acnkern_set_depth (acnguy->search, 4);              /* default level */
  acnkern_set_callback_count (acnguy->search, 10000); /* default count */

#ifdef WIN32
  if ((handle = CreateThread (NULL,
                    0,
                    (LPTHREAD_START_ROUTINE) acnguy_process_input,
                    acnguy,
                    0,
                    &pth)) == NULL)
      {
        fprintf (acnguy->error_fp, "Start input process thread failed.\n");
        exit (2);
      }

  do
    {
      if (GetExitCodeThread (handle, &ret) == 0)
        {
          break;
        }
    }
  while (ret == STILL_ACTIVE);
#else
    if (pthread_create (&pth, 
                        NULL, 
                        (void *(*) (void *)) acnguy_process_input, 
                        acnguy) != 0)
      {
        fprintf (acnguy->error_fp, "Start input process thread failed.\n");
        exit (2);
      }

    pthread_join (pth, NULL);
#endif

  return 0;
}

static int
acnguy_process_ucci (struct acnguy *acnguy, ucci_cmd *cmd)
{
  fputs ("id name acnguy 0.0.1\n", acnguy->write_fp);
  fputs ("id author Alf <naihe2010@126.com>\n", acnguy->write_fp);
  fputs ("ucciok\n", acnguy->write_fp); 
  fflush (acnguy->write_fp);
  return 0;
};

static int
acnguy_process_isready (struct acnguy *acnguy, ucci_cmd *cmd)
{
  fputs ("readyok\n", acnguy->write_fp);
  fflush (acnguy->write_fp);
  return 0;
};

static int
acnguy_process_setoption (struct acnguy *acnguy, ucci_cmd *cmd)
{
  return 0;
};

static int
acnguy_process_position (struct acnguy *acnguy, ucci_cmd *cmd)
{
  size_t i, moveind = 0;
  struct xq_move move[1];

  if (strcmp (cmd->args[0], "fen") == 0)
    {
      char string[0x100];

      sprintf (string, "%s %s %s %s %s %s", 
               cmd->args[1],
               cmd->args[2],
               cmd->args[3],
               cmd->args[4],
               cmd->args[5],
               cmd->args[6]);
      fen_from_string (acnguy->fen, string);

      if (cmd->arg_count > 7 && strcmp (cmd->args[7], "moves") == 0)
        {
          moveind = 8;
        }
    }
  else if (strcmp (cmd->args[0], "startpos") == 0)
    {
      memcpy (acnguy->fen->pan, default_pan, 10 * 9 * sizeof (char));
      if (cmd->arg_count > 1 && strcmp (cmd->args[1], "moves") == 0)
        {
          moveind = 2;
        }
    }
  else
    {
      return -1;
    }

  acnkern_set_pan (acnguy->search, acnguy->fen->pan);

  if (moveind)
    {
      for (i = moveind; i < cmd->arg_count; ++ i)
        {
          iccs_to_move (cmd->args[i], move);
          acnkern_set_move (acnguy->search, move);
        }
    }

  return 0;
};

static int
acnguy_process_banmoves (struct acnguy *acnguy, ucci_cmd *cmd)
{
  return 0;
};

static int
acnguy_process_go (struct acnguy *acnguy, ucci_cmd *cmd)
{
  char buf[0x100];
  struct xq_move *move;

  move = acnkern_go (acnguy->search, 0);

  if (move)
    {
      memcpy (buf, "bestmove ", 9);

      iccs_from_move (buf + 9, move);
      buf[13] = '\n';
      buf[14] = '\0';

      fputs (buf, acnguy->write_fp);
      fflush (acnguy->write_fp);

      return 0;
    }
  else
    {
      return -1;
    }
};

static int
acnguy_process_ponderhit (struct acnguy *acnguy, ucci_cmd *cmd)
{
  return 0;
};

static int
acnguy_process_stop (struct acnguy *acnguy, ucci_cmd *cmd)
{
  return 0;
};

static int
acnguy_process_probe (struct acnguy *acnguy, ucci_cmd *cmd)
{
  return 0;
};

static int
acnguy_process_quit (struct acnguy *acnguy, ucci_cmd *cmd)
{
  fputs ("bye\n", acnguy->write_fp);
  fflush (acnguy->write_fp);
  exit (0);
  return 0;
};

static int (*cmdcbs[]) (struct acnguy *, ucci_cmd *) = {
    acnguy_process_ucci,
    NULL, /* id */
    NULL, /* option */
    NULL, /* ucciok */
    acnguy_process_isready,
    NULL, /* readyok */
    acnguy_process_setoption,
    acnguy_process_position,
    acnguy_process_banmoves,
    acnguy_process_go,
    NULL, /* info */
    acnguy_process_ponderhit,
    acnguy_process_stop,
    NULL, /* bestmove */
    NULL, /* nobestmove */
    acnguy_process_probe,
    NULL, /* pophash */
    acnguy_process_quit,
    NULL, /* bye */
};

#define cmdcbs_count    (sizeof(cmdcbs)/sizeof(cmdcbs[0]))

static int 
acnguy_process (struct acnguy *acnguy, char *string)
{
  ucci_cmd cmd[1];

  if (ucci_cmd_from_string (cmd, string) < 0)
    {
      return -1;
    }

  if (cmdcbs[cmd->cmd])
    {
      cmdcbs[cmd->cmd] (acnguy, cmd);
    }

  return 0;
}

static int 
acnguy_process_input (struct acnguy *acnguy)
{
  char inbuf[ACNGUY_INBUF_SIZE];
    
  while (fgets (inbuf, sizeof inbuf, acnguy->read_fp))
    {
      acnguy_process (acnguy, inbuf);
    }

  return 1;
}

static int 
acnguy_engine_callback (struct acnguy *acnguy)
{
  return 1;
}
