/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE proxysrv.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/08/27 10:26:38 Alf*/

#include "proxysrv.h"
#include "status.h"
#include "packet.h"
#include "log.h"

#include <gtk/gtk.h>

#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#else
#include <winsock2.h>
#define socklen_t int
#endif

static int set_allfd (proxy_srv *, fd_set *);
static int proc_allfd (proxy_srv *, fd_set *);

static void dclient_clear (proxy_srv *, dclient_t *);

static int dclient_process (proxy_srv *, dclient_t *, const char *, int);

static int process_action_request (proxy_srv *, dclient_t *, packet_t *,
                                   packet_t *);
static int process_get_userlist (proxy_srv *, dclient_t *, packet_t *,
                                 packet_t *);

proxy_srv *
proxysrv_new (int max, unsigned short port)
{
  proxy_srv *srv;
  int daemon, i;
  struct sockaddr_in loc[1];

  if (max <= 0 || max > 32)
    {
      acc_err (_("Proxy client is too little\n"));
      return NULL;
    }

  daemon = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (daemon < 0)
    {
      acc_err (_("socket error\n"));
      return NULL;
    }

  i = 1;
#ifndef WIN32
  ioctl (daemon, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (i));
#else
  /* setsockopt (daemon, SOL_SOCKET, SO_RESUEADDR, &i, sizeof (i)); */
#endif

  loc->sin_family = AF_INET;
  loc->sin_addr.s_addr = INADDR_ANY;
  loc->sin_port = htons (port);
  memset (loc->sin_zero, 0, 8);
  if (bind (daemon, (struct sockaddr *) loc, sizeof (struct sockaddr)) < 0)
    {
      acc_err (_("bind error\n"));
      close (daemon);
      return NULL;
    }

  if (listen (daemon, 10) < 0)
    {
      acc_err (_("listen error\n"));
      close (daemon);
      return NULL;
    }

  srv = calloc (1, sizeof (proxy_srv));
  if (srv == NULL)
    {
      acc_err (_("Memory error\n"));
      close (daemon);
      return NULL;
    }

  srv->fd = daemon;

  return srv;
}

void
proxysrv_delete (proxy_srv * srv)
{
  close (srv->fd);
  free (srv);
}

int
proxysrv_start (proxy_srv * srv)
{
  fd_set rset[1];
  struct timeval tv[1];
  int max, ret;

  srv->running = 1;

  while (srv->running)
    {
      max = set_allfd (srv, rset);
      if (max <= 0)
        {
          continue;
        }

      tv->tv_sec = 2;
      tv->tv_usec = 0;
      ret = select (max, rset, NULL, NULL, tv);
      if (ret <= 0)
        {
          continue;
        }

      proc_allfd (srv, rset);
    }

  return 0;
}

void
proxysrv_stop (proxy_srv * srv)
{
  srv->running = 0;
}

static int
set_allfd (proxy_srv * srv, fd_set * set)
{
  dclient_t *cli;
  int i, max;

  FD_ZERO (set);

  FD_SET (srv->fd, set);
  for (max = srv->fd, i = 0; i < srv->cur; ++i)
    {
      cli = srv->clients + i;

      if (cli->fd <= 0)
        {
          continue;
        }

      FD_SET (cli->fd, set);
      if (cli->fd > max)
        {
          max = cli->fd;
        }
    }

  return max + 1;
}

static int
proc_allfd (proxy_srv * srv, fd_set * set)
{
  dclient_t *cli;
  char buf[0x2000];
  int i, ret;

  if (FD_ISSET (srv->fd, set))
    {
      struct sockaddr_in rmt[1];
      socklen_t len;
      int fd;

      len = sizeof (struct sockaddr);
      fd = accept (srv->fd, (struct sockaddr *) rmt, &len);
      acc_debug ("new connection from %s:%d\n", inet_ntoa (rmt->sin_addr),
                 ntohs (rmt->sin_port));
      if (fd > 0)
        {
          for (i = 0; i < srv->cur; ++i)
            {
              cli = srv->clients + i;
              if (cli->fd <= 0)
                {
                  break;
                }
            }

          cli = srv->clients + i;
          cli->fd = fd;
          cli->ip = rmt->sin_addr.s_addr;
          cli->port = rmt->sin_port;

          if (i == srv->cur)
            {
              srv->cur++;
            }
        }
    }

  for (i = 0; i < srv->cur; ++i)
    {
      cli = srv->clients + i;

      if (cli->fd <= 0)
        {
          continue;
        }

      if (FD_ISSET (cli->fd, set))
        {
          ret = read (cli->fd, buf, sizeof buf);
          if (ret <= 0)
            {
              dclient_clear (srv, cli);
            }
          else
            {
              dclient_process (srv, cli, buf, ret);
            }
        }
    }

  return 0;
}

static int
dclient_process (proxy_srv * srv, dclient_t * cli, const char *buf, int siz)
{
  packet_t req[1] = { {0} }, res[1] =
  {
    {
      0
    }
  };
  char sendbuf[0x2000];
  int sendsiz;

  if (packet_decode (buf, siz, req) < 0)
    {
      acc_debug ("%d bytes decode error\n", siz);
      return -1;
    }

  switch (req->type)
    {
    case ACP_REG:
      break;

    case ACP_ACTION_REQUEST:
      process_action_request (srv, cli, req, res);

    case ACP_GET_USERLIST:
      process_get_userlist (srv, cli, req, res);
      break;

    default:
      break;
    }

  if ((sendsiz = packet_encode (res, sendbuf, sizeof sendbuf)) <= 0)
    {
      acc_debug ("no response\n");
    }

  if (write (cli->fd, sendbuf, sendsiz) < sendsiz)
    {
      acc_debug ("client from %s:%d disconnect\n",
                 inet_ntoa (*(struct in_addr *) &cli->ip), ntohs (cli->port));
    }

  packet_free (req);
  packet_free (res);

  return 0;
}

static void
dclient_clear (proxy_srv * srv, dclient_t * cli)
{
  if (cli->fd > 0)
    {
      close (cli->fd);
      cli->fd = 0;
    }
}

static int
process_action_request (proxy_srv * srv, dclient_t * cli, packet_t * req,
                        packet_t * res)
{
  char mes[0x100];
  GtkWidget *dia;
  gint ret;

  g_snprintf (mes, sizeof mes,
              _("Client [%s] from [%s:%d] want to fight with you.\n"), "test",
              inet_ntoa (*(struct in_addr *) &cli->ip), ntohs (cli->port));
  dia =
    gtk_dialog_new_with_buttons (_("Request:"), NULL,
                                 GTK_DIALOG_DESTROY_WITH_PARENT,
                                 GTK_MESSAGE_INFO, GTK_BUTTONS_CLOSE, mes);
  ret = gtk_dialog_run (GTK_DIALOG (dia));
  if (ret == GTK_RESPONSE_ACCEPT)
    {
    }

  gtk_widget_destroy (dia);

  return 0;
}

static int
process_get_userlist (proxy_srv * srv, dclient_t * cli, packet_t * req,
                      packet_t * res)
{
  dclient_t *c;
  int i;
  char buffer[0x1000];
  int len, onelen;
  static const char *status_phrase[] =
  {
    "Connect",
    "Ready",
    "Fighting",
  };

  for (len = i = 0; i < srv->cur; ++i)
    {
      c = srv->clients + i;
      if (c->id)
        {
          onelen = snprintf (buffer + len, sizeof buffer - len,
                             "%s:%s:%s\n",
                             c->id, "Xiang Qi", status_phrase[c->status]);
          len += onelen;
        }
    }

  len +=
    snprintf (buffer + len, sizeof buffer - len, "%s:%s:%s\n", "", "Xiang Qi",
              status_phrase[READY]);
  acc_debug ("response %d[%s]\n", len, buffer);

  if (len > 0)
    {
      res->elem = packet_elem_new (ACP_GET_USERLIST, len, buffer);
      res->type = 1;
    }
  else
    {
      res->elem = NULL;
      res->type = 0;
    }

  return 0;
}
