﻿/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE qipu.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/08/12 11:16:14 Alf*/

#include "qipu.h"
#include "log.h"
#include "acnkern.h"
#include "fen.h"
#include "zgxq.h"

#include <glib/gstdio.h>

#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

struct tv_node
{
  gchar *token;
  gchar *value;
};

struct qipu_s {
  GData * tv_list;

  fen_t fen[1];

  GList * move_cur;
  GList * move_list;
};

static int qp_load_pgn_fp (qipu_t *, FILE *);

static int desc_get_number (const char *, int *);

qipu_t *
qipu_new ()
{
  qipu_t * qipu;

  qipu = calloc (1, sizeof (qipu_t));
  if (qipu == NULL)
    {
      acc_err (_ ("Memory error.\n"));
      return NULL;
    }

  g_datalist_init (&qipu->tv_list);

  return qipu;
}

qipu_t *
qipu_new_from_file (const char *file)
{
  qipu_t * qipu;

  qipu = qipu_new ();
  if (qipu == NULL)
    {
      return NULL;
    }

  if (qipu_load (qipu, file) < 0)
    {
      qipu_destroy (qipu);
      return NULL;
    }
  return qipu;
}

int
qipu_load (qipu_t *qipu, const char *file)
{
  FILE *fp;
  size_t namelen;
  int ret;

  fp = fopen (file, "r");
  if (fp == NULL)
    {
      acc_err (_ ("open file '%s' error: %s\n"), file, strerror (errno));
      return -1;
    }

  namelen = strlen (file);
  if (g_strcasecmp (file + namelen - 4, ".pgn") == 0)
    {
      ret = qp_load_pgn_fp (qipu, fp);
    }
  else
    {
      ret = -1;
    }

  fclose (fp);

  return ret;
}

int
qipu_save (qipu_t *qipu, const char *file)
{
  FILE *fp;
  const char *text;
  gchar *ufile;
  struct xq_move mover[1], moveb[1];
  int ret, id;
  char mes[0x100], pan[10][9];

  ufile = g_locale_from_utf8 (file, -1, NULL, NULL, NULL);
  if (ufile)
    {
      fp = fopen (ufile, "a");
      g_free (ufile);
    }
  else
    {
      fp = fopen (file, "a");
    }

  if (fp == NULL)
    {
      acc_err ( _("Open file '%s' error: %s\n"), file, strerror (errno));
      return -1;
    }

  text = qipu_get_info (qipu, "Event");
  if (text)
    {
      fprintf (fp, "[Event \"%s\"]\n", text);
    }

  text = qipu_get_info (qipu, "Date");
  if (text)
    {
      fprintf (fp, "[Date \"%s\"]\n", text);
    }

  text = qipu_get_info (qipu, "Site");
  if (text)
    {
      fprintf (fp, "[Site \"%s\"]\n", text);
    }

  text = qipu_get_info (qipu, "Red");
  if (text)
    {
      fprintf (fp, "[Red \"%s\"]\n", text);
    }

  text = qipu_get_info (qipu, "Black");
  if (text)
    {
      fprintf (fp, "[Black \"%s]\"\n", text);
    }

  text = qipu_get_info (qipu, "FEN");
  if (text)
    {
      fprintf (fp, "[FEN %s]\n", text);
    }
  else
    {
      char string[0x100];
      fen_to_string (qipu->fen, string, sizeof string);
      fprintf (fp, "[FEN %s]\n", string);
    }

  text = qipu_get_info (qipu, "Commet");
  if (text)
    {
      fprintf (fp, "\n{%s}\n\n", text);
    }

  memcpy (pan, qipu->fen->pan, 10 * 9 * sizeof (char));
  for (id = 0, ret = qipu_move_first (qipu, mover);
       ret == 0;
       ++ id, ret = qipu_move_next (qipu, NULL, mover))
    {
      fprintf (fp, "%d. ", id);

      mover->isred = 1;
      move_to_desc (pan, mover, mes, sizeof mes);
      pan[mover->toy][mover->tox] = pan[mover->fromy][mover->fromx];
      pan[mover->fromy][mover->fromx] = 0;
      fprintf (fp, " %s ", mes);

      qipu_move_next (qipu, NULL, moveb);

      moveb->isred = 0;
      move_to_desc (pan, moveb, mes, sizeof mes);
      pan[moveb->toy][moveb->tox] = pan[moveb->fromy][moveb->fromx];
      pan[moveb->fromy][moveb->fromx] = 0;
      fprintf (fp, " %s ", mes);

      fprintf (fp, "\n");
    }

  fclose (fp);

  return 0;
}

void
qipu_destroy (qipu_t *qipu)
{
  g_datalist_clear (&qipu->tv_list);
  free (qipu);
}

#ifdef _DEBUG
static void
qp_output_tv_debug (GQuark id, gpointer data)
{
  acc_debug ("[%s]->[%s].\n", g_quark_to_string (id), (gchar *) data);
}
#endif

static int
qp_load_pgn_fp (qipu_t *qipu, FILE *fp)
{
  char line[0x500], pan[10][9], *b, *e, *m; 
  const char *fen;
  int incommet, isred;
#ifdef _DEBUG
  char move_str[0x100];
#endif

  while (fgets (line, sizeof line, fp) != NULL)
    {
      b = strchr (line, '[');
      if (b == NULL)
        {
          break;
        }

      do
        {
          ++ b;
        }
      while (isspace (*b));

      if (*b == '\0')
        {
          acc_warn ("parser line: '%s' error.\n", line);
          break;
        }

      m = strchr (b, ' ');
      if (m == NULL)
        {
          m = strchr (b, '\t');
        }
      if (m == NULL)
        {
          acc_warn ("parser line: '%s' error.\n", line);
          break;
        }

      *m = '\0';                /* token end */

      do
        {
          ++ m;
        }
      while (isspace (*m));

      e = strchr (m, ']');
      if (e == NULL)
        {
          acc_warn ("parser line: '%s' error.\n", line);
          break;
        }

      do
        {
          *e = '\0';
          -- e;
        }
      while (isspace (*e));

      if (*m == '"' || *m == '\'')
        {
          ++ m;
        }
      if (*e == '"' || *e == '\'')
        {
          *e = '\0';
        }

      qipu_set_info (qipu, b, m);
    }

#ifdef _DEBUG
  g_datalist_foreach (&qipu->tv_list, (GDataForeachFunc) qp_output_tv_debug, NULL);
#endif

  fen = qipu_get_info (qipu, "FEN");
  if (fen == NULL)
    {
      fen = qipu_get_info (qipu, "fen");
    }
  if (fen)
    {
      fen_from_string (qipu->fen, fen);
      memcpy (pan, qipu->fen->pan, 10 * 9 * sizeof (char));
    }
  else
    {
      memcpy (pan, default_pan, 10 * 9 * sizeof (char));
    }

  incommet = 0;
  isred = 1;

  do 
    {
      //acc_debug ("get line: [%s]\n", line);

      b = line;

      if (incommet)
        {
          b = strchr (b, '}');
          if (b == NULL)
            {
              continue;
            }
          else
            {
              incommet = 0;
              ++ b;
            }
        }

      while (isspace (*b)) 
        {
          ++ b;
        }

      if (*b == '{')
        {
          b = strchr (b, '}');
          if (b == NULL)
            {
              incommet = 1;
              continue;
            }
          else
            {
              ++ b;
            }
        }

      while (isspace (*b))
        {
          ++ b;
        }
      if (*b == '\0')
        {
          continue;
        }

      while (*b != '\0')
        {
          struct xq_move move[1];
          int offset;

          /* strip blank or numbers and . */
          while (isspace (*b))
            {
              ++ b;
            }
          while (desc_get_number (b, &offset) >= 0)
            {
              b += offset;
            }
          if (*b == '.')
            {
              ++ b;
            }
          while (isspace (*b))
            {
              ++ b;
            }
          if (*b == '\0')
            {
              break;
            }

          move->isred = isred;

          isred = isred ? 0: 1;

#ifdef _DEBUG
          acc_debug ("\n\nbegin move\n");
          output_pan (pan);
#endif
          offset = desc_to_move (pan, b, move);
          if (offset <= 0)
            {
              break;
            }
          b += offset;
#ifdef _DEBUG
          move_to_desc (pan, move, move_str, sizeof move_str);
          acc_debug ("desc: %s\n", move_str);
#endif
          pan[move->toy][move->tox] = pan[move->fromy][move->fromx];
          pan[move->fromy][move->fromx] = 0;
          qipu_append_move (qipu, move);
#ifdef _DEBUG
          output_pan (pan);
          acc_debug ("after move\n\n");
#endif
        }

      //acc_debug ("get not commet: [%s]\n", b);
    }
  while (fgets (line, sizeof line, fp) != NULL);

  return 0;
}

int
qipu_set_info (qipu_t *qipu, const char *token, const char *value)
{
  gpointer data;

  data = g_strdup (value);
  g_datalist_set_data_full (&qipu->tv_list, token, data, free);
  return 0;
}

const char *
qipu_get_info (qipu_t *qipu, const char *token)
{
  gpointer data;

  data = g_datalist_get_data (&qipu->tv_list, token);
  return data;
}

int 
qipu_get_fen (qipu_t * qipu, fen_t * fen)
{
  memcpy (fen, qipu->fen, sizeof (fen_t));
  return 0;
}

int 
qipu_set_fen (qipu_t * qipu, fen_t * fen)
{
  memcpy (qipu->fen, fen, sizeof (fen_t));
  return 0;
}

int
desc_to_move (char pan[10][9], const char *in, struct xq_move *move)
{
  int front, step, zi, offset, i, j;
  gchar *udesc; 
  const char *from, *desc;
  size_t chlen = strlen ("前");      /* maybe bug ??? */

  udesc = g_locale_from_utf8 (in, -1, NULL, NULL, NULL);
  if (udesc)
    {
      desc = udesc;
    }
  else
    {
      desc = in;
    }

  from = desc;

  if (strncmp (desc, "前", chlen) == 0)
    {
      front = 1;
      desc += chlen;
    }
  else if (strncmp (desc, "后", chlen) == 0)
    {
      front = -1;
      desc += chlen;
    }
  else
    {
      front = 0;
    }

  for (zi = 0, i = (move->isred ? R_SHUAI : B_SHUAI); i <= B_BING; i += 2)
    {
      if (strncmp (desc, zi_chinese[i], chlen) == 0)
        {
          zi = i;
          break;
        }
    }

  if (zi == 0)
    {
      acc_err ("Can't find this qizi.\n");
      return -1;
    }
  acc_debug ("get qizi: %s[%s]\n", move->isred ? "红": "黑", zi_chinese[(int)zi]);

  desc += chlen;

  if (front != 0)
    {
      for (i = 0; i < 10; ++ i)
        {
          for (j = 0; j < 9; ++ j)
            {
              if (pan[i][j] == zi)
                {
                  if ((front > 0 && move->isred)
                      || (front < 0 && !move->isred))
                    {
                      move->fromy = i;
                      move->fromx = j;
                      goto find_front;
                    }
                  else if (front != 0)
                    {
                      front = 0;
                      continue;
                    }
                  else
                    {
                      move->fromy = i;
                      move->fromx = j;
                      goto find_front;
                    }
                }
            }
        }

find_front:

      if (i == 10)
        {
          acc_warn ( "Can't find this qizi.\n");
          return -1;
        }
    }
  else
    {
      move->fromx = desc_get_number (desc, &offset);
      acc_debug ("get number: %d\n", move->fromx);
      if (move->isred)
        {
          move->fromx = 9 - move->fromx;
        }
      else
        {
          move->fromx = move->fromx - 1;
        }
      acc_debug ("get move->fromx: %d\n", move->fromx);
      desc += offset;

      for (j = 0; j <= 9; ++ j)
        {
          if (pan[j][move->fromx] == zi)
            {
              move->fromy = j;
              break;
            }
        }
      if (j > 9)
        {
          acc_debug ("Can't find this qizi.\n");
          return -1;
        }
    }

  step = desc_get_number (desc + chlen, &offset);

  if (strncmp (desc, "平", chlen) == 0)
    {
      move->toy = move->fromy;
      move->tox = step;
      if (move->isred)
        {
          move->tox = 9 - move->tox;
        }
      else
        {
          move->tox = move->tox - 1;
        }
    }
  else if (strncmp (desc, "进", chlen) == 0)
    {
      move->tox = step;
      if (move->isred)
        {
          move->tox = 9 - move->tox;
        }
      else
        {
          move->tox = move->tox - 1;
        }

      if (zi == R_MA)
        {
          if (move->tox - move->fromx == 2
              || move->tox - move->fromx == -2)
            {
              move->toy = move->fromy - 1;
            }
          else
            {
              move->toy = move->fromy - 2;
            }
        }
      else if (zi == B_MA)
        {
          if (move->tox - move->fromx == 2
              || move->tox - move->fromx == -2)
            {
              move->toy = move->fromy + 1;
            }
          else
            {
              move->toy = move->fromy + 2;
            }
        }
      else if (zi == R_SHI)
        {
          move->toy = move->fromy - 1;
        }
      else if (zi == B_SHI)
        {
          move->toy = move->fromy + 1;
        }
      else if (zi == R_XIANG)
        {
          move->toy = move->fromy - 2;
        }
      else if (zi == B_XIANG)
        {
          move->toy = move->fromy + 2;
        }
      else
        {
          move->tox = move->fromx;

          if (move->isred)
            {
              move->toy = move->fromy - step;
            }
          else
            {
              move->toy = move->fromy + step;
            }
        }
    }
  else if (strncmp (desc, "退", chlen) == 0)
    {
      move->tox = step;
      if (move->isred)
        {
          move->tox = 9 - move->tox;
        }
      else
        {
          move->tox = move->tox - 1;
        }

      if (zi == R_MA)
        {
          if (move->tox - move->fromx == 2
              || move->tox - move->fromx == -2)
            {
              move->toy = move->fromy + 1;
            }
          else
            {
              move->toy = move->fromy + 2;
            }
        }
      else if (zi == B_MA)
        {
          if (move->tox - move->fromx == 2
              || move->tox - move->fromx == -2)
            {
              move->toy = move->fromy - 1;
            }
          else
            {
              move->toy = move->fromy - 2;
            }
        }
      else if (zi == R_SHI)
        {
          move->toy = move->fromy + 1;
        }
      else if (zi == B_SHI)
        {
          move->toy = move->fromy - 1;
        }
      else if (zi == R_XIANG)
        {
          move->toy = move->fromy + 2;
        }
      else if (zi == B_XIANG)
        {
          move->toy = move->fromy - 2;
        }
      else
        {
          move->tox = move->fromx;

          if (move->isred)
            {
              move->toy = move->fromy + step;
            }
          else
            {
              move->toy = move->fromy - step;
            }
        }
    }

  desc += chlen;
  desc += offset;

  acc_debug ("transer [%s] to [%d: %d:%d->%d:%d]\n", 
             from, zi, 
             move->fromx, move->fromy, move->tox, move->toy);
  if (udesc)
    {
      g_free (udesc);
    }

  return desc - from;
}

ssize_t
move_to_desc (char pan[10][9], struct xq_move *move, char *out, size_t len)
{
  int i, front, to, ret, zi;
  char desc[0x100], *p;
  gchar *udesc;

  for (zi = pan[move->fromy][move->fromx], front = 0, i = 0; 
       i < 9; 
       ++ i)
    {
      if (i == move->fromy)
        {
          continue;
        }
      if (pan[i][move->fromx] == zi)
        {
          front = i - move->fromy;
        }
    }

  p = desc;
  *p = 0;

  if ((front > 0 && move->isred)
      || (front < 0 && (!move->isred)))
    {
      strcpy (p, "前");
    }
  else if (front != 0)
    {
      strcpy (p, "后");
    }
  strcat (p, zi_chinese[zi]);

  if (front == 0)
    {
      if (move->isred)
        {
          strcat (p, number_chinese_str[9 - move->fromx]);
        }
      else
        {
          strcat (p, number_b_str[move->fromx + 1]);
        }
    }

  if ((move->isred && move->toy < move->fromy)
      || (!move->isred && move->toy > move->fromy)
  )
    {
      strcat (p, "进");
      if (zi == B_MA
          || zi == R_MA
          || zi == B_SHI
          || zi == R_SHI
          || zi == B_XIANG
          || zi == R_XIANG
      )
        {
          to = move->isred ? 9 - move->tox: move->tox + 1;
        }
      else
        {
          to = move->toy - move->fromy;
        }
    }
  else if (move->fromy != move->toy)
    {
      strcat (p, "退");
      if (zi == B_MA
          || zi == R_MA
          || zi == B_SHI
          || zi == R_SHI
          || zi == B_XIANG
          || zi == R_XIANG
      )
        {
          to = move->isred ? 9 - move->tox: move->tox + 1;
        }
      else
        {
          to = move->toy - move->fromy;
        }
    }
  else
    {
      strcat (p, "平");
      to = 0;
    }

  if (to < 0)
    {
      to = - to;
    }

  if (to > 0)
    {
      if (move->isred)
        {
          strcat (p, number_chinese_str[to]);
        }
      else
        {
          strcat (p, number_b_str[to]);
        }
    }
  else
    {
      if (move->isred)
        {
          strcat (p, number_chinese_str[9 - move->tox]);
        }
      else
        {
          strcat (p, number_b_str[move->tox + 1]);
        }
    }

  udesc = g_locale_to_utf8 (desc, -1, NULL, NULL, NULL);
  if (udesc)
    {
      ret = g_snprintf (out, len, "%s", udesc);
      g_free (udesc);
    }
  else
    {
      ret = g_snprintf (out, len, "%s", desc);
    }

  return ret;
}

int 
qipu_append_move (qipu_t *qipu, struct xq_move *move)
{
  struct xq_move *nmove;

  nmove = malloc (sizeof (struct xq_move));
  if (nmove == NULL)
    {
      acc_err (_ ("Memory error.\n"));
      return -1;
    }

  memcpy (nmove, move, sizeof (struct xq_move));
  qipu->move_list = g_list_append (qipu->move_list, nmove);

  return 0;
}

int
qipu_move_first (qipu_t *qipu, struct xq_move *move)
{
  struct xq_move *node;

  qipu->move_cur = g_list_first (qipu->move_list);
  if (qipu->move_cur == NULL)
    {
      return -1;
    }

  node = qipu->move_cur->data;
  if (node == NULL)
    {
      return -1;
    }

  memcpy (move, node, sizeof (struct xq_move));
  return 0;
}

int
qipu_move_next (qipu_t *qipu, const struct xq_move *move, struct xq_move *next)
{
  struct xq_move *node;
  GList *cur;

  if (move)
    {
      cur = g_list_find (qipu->move_list, move);
    }
  else
    {
      cur = qipu->move_cur;
    }

  if (cur == NULL)
    {
      return -1;
    }

  qipu->move_cur = g_list_next (cur);
  if (qipu->move_cur == NULL)
    {
      return -1;
    }

  node = qipu->move_cur->data;
  if (node == NULL)
    {
      return -1;
    }

  memcpy (next, node, sizeof (struct xq_move));
  return 0;
}

static int
desc_get_number (const char *str, int *numlen)
{
  int i, num, lblen, bblen, cblen;

  num = -1;
  if (numlen)
    {
      *numlen = 0;
    }

  lblen = strlen ("1");
  bblen = strlen ("１");
  cblen = strlen ("一");

  for (i = 0; i <= 9; ++ i)
    {
      if (strncmp (str, number_str[i], lblen) == 0)
        {
          num = i;
          if (numlen)
            {
              *numlen = lblen;
            }
          break;
        }

      if (strncmp (str, number_b_str[i], bblen) == 0)
        {
          num = i;
          if (numlen)
            {
              *numlen = bblen;
            }
          break;
        }

      if (strncmp (str, number_chinese_str[i], cblen) == 0)
        {
          num = i;
          if (numlen)
            {
              *numlen = cblen;
            }
          break;
        }
    }

  return num;
}

#ifdef QIPU_DEBUG
int
main (int argc, char *argv[])
{
  qipu_t *qipu;

  qipu = qipu_new_from_file (argv[1]);

  return 0;
}
#endif
