/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE session.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/09/04 23:50:14 Alf*/

#include "session.h"
#include "log.h"
#include "zgxq.h"
#include "fen.h"
#include "iccs.h"

#ifdef WIN32
#else
#include <unistd.h>
#endif

#include <glib.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

struct session_s
{
  char engine[PATH_MAX];

#ifdef WIN32
  HANDLE hRead;
  HANDLE hWrite;

  char buffer[0x1000];
  size_t buflen;
#else
  int fd;
  FILE *fpin;
  FILE *fpout;
#endif
};

static char *session_gets (session_t *, char *, size_t);
static int session_puts (session_t *, const char *);

#ifndef WIN32
static int
ppopen (const char * command)
{
  int fds[2], pid;

  if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, fds) < 0)
    {
      acc_err ("sockerpair error: %s\n", strerror (errno));
      return -1;
    }

  if ((pid = fork()) < 0)
    {
      acc_err ("fork error: %s\n", strerror (errno));
      close(fds[0]);
      close(fds[1]);
      return -1;
    }

  if (pid > 0)
    {
      if (close(fds[1]) < 0)
        {
          return -1;
        }
      return fds[0];
    }
  else
    {
      if (close(fds[0]) < 0)
        {
          exit (1);
        }
      if (dup2(fds[1], 0) < 0 || dup2(fds[1], 1) < 0)
        {
          exit(1);
        }
      if (close(fds[1]) < 0)
        {
          exit(1);
        }

      execl("/bin/sh", "sh", "-c", command, NULL);
      exit(1);
    }
}
#endif

session_t *
session_new (const char *name)
{
#ifdef WIN32
  char command[1024];
  SECURITY_ATTRIBUTES sa;
  HANDLE hRead, hWrite;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
#endif
  session_t *ses;

  ses = calloc (1, sizeof (session_t));
  if (ses == NULL)
    {
      acc_err (_("Memory error\n"));
      return NULL;
    }

#ifdef WIN32
  sa.nLength = sizeof (SECURITY_ATTRIBUTES);
  sa.lpSecurityDescriptor = NULL;
  sa.bInheritHandle = TRUE;
  if (!CreatePipe(&hRead, &ses->hWrite, &sa, 0))
    {
      free (ses);
      return NULL;
    }
  if (!CreatePipe (&ses->hRead, &hWrite, &sa, 0))
    {
      CloseHandle (hRead);
      CloseHandle (ses->hWrite);
      free (ses);
      return NULL;
    }

  g_snprintf (command, sizeof command, "%s", name);

  si.cb = sizeof(STARTUPINFO);
  GetStartupInfo(&si);
  si.hStdOutput = hWrite;
  si.hStdInput = hRead;
  si.wShowWindow = SW_HIDE;
  si.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;

  if (!CreateProcess(NULL, command, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi))
    {
      CloseHandle(hWrite);
      CloseHandle(hRead);
      CloseHandle (ses->hWrite);
      CloseHandle (ses->hRead);
      free (ses);
      return NULL;
    }
  CloseHandle (hWrite);
  CloseHandle (hRead);

#else
  ses->fd = ppopen (name);
  if (ses->fd < 0)
    {
      free (ses);
      return NULL;
    }

  ses->fpin = fdopen (ses->fd, "r");
  ses->fpout = fdopen (ses->fd, "w");

  if (ses->fpin == NULL
      || ses->fpout == NULL)
    {
      if (ses->fpin)
        {
          fclose (ses->fpin);
        }
      close (ses->fd);
      free (ses);
      return NULL;
    }
#endif

  return ses;
}

void
session_delete (session_t * ses)
{
#ifdef WIN32
  CloseHandle (ses->hRead);
  CloseHandle (ses->hWrite);
#else
  fclose (ses->fpin);
  fclose (ses->fpout);
#endif
  free (ses);
}

const char *
session_desc (session_t * ses)
{
  static char buf[0x200] = { "UNKNOWN" };

  snprintf (buf, sizeof buf, "Engine");

  return buf;
}

struct xq_move *
session_get_move (session_t * ses, fen_t *fen, GList *movelist, int depth, int milsecs)
{
  static struct xq_move move[1];
  char buf[0x1000], string[0x100], iccs[6];
  GList *cur;

  strcpy (buf, "position fen ");
  fen_to_string (fen, string, sizeof string);
  strcat (buf, string);
  if (fen->half_count)
    {
      strcat (buf, " moves ");
      for (cur = movelist; cur != NULL; cur = g_list_next (cur))
        {
          iccs_from_move (iccs, cur->data);
          iccs[4] = ' ';
          iccs[5] = '\0';
          strcat (buf, iccs);
        }
    }
  strcat (buf, "\n");

  session_puts (ses, buf);
  acc_debug ("send '%s' to engine\n", buf);

  sprintf (buf, "go depth %d\n", depth);
  session_puts (ses, buf);
  acc_debug ("send '%s' to engine\n", buf);

  while (session_gets (ses, buf, sizeof buf) != NULL)
    {
      acc_debug ("get '%s' from engine\n", buf);
      if (strncmp (buf, "bestmove ", 9) == 0)
        {
          iccs_to_move (buf + 9, move);
          break;
        }
    }

  return move;
}

int
session_start (session_t * ses)
{
  char buf[0x100];

  session_puts (ses, "ucci\n");
  while (session_gets (ses, buf, sizeof buf) != NULL)
    {
      acc_debug ("get '%s' from engine\n", buf);
      if (strncmp (buf, "ucciok", 6) == 0)
        {
          break;
        }
    }

  return 0;
}

int
session_hui (session_t * ses)
{
  return 0;
}

int
session_stop (session_t * ses)
{
  return 0;
}

static char *
session_gets (session_t *ses, char *in, size_t cap)
{
#ifdef WIN32
  size_t i, len;
  DWORD Bytes;

  while (1)
    {
      if (ses->buflen > 0)
        {
          for (i = 0;
               i < ses->buflen;
               ++ i)
            {
              if (ses->buffer[i] == '\n')
                {
                  break;
                }
            }

          if (i < ses->buflen)
            {
              len = i + 1;
              memcpy (in, ses->buffer, len);
              in[len] = '\0';

              ses->buflen -= len;
              memmove (ses->buffer, ses->buffer + len, ses->buflen);

              return in;
            }
        }

      ReadFile(ses->hRead, ses->buffer + ses->buflen, sizeof (ses->buffer) - ses->buflen, &Bytes, NULL);
      if (Bytes <= 0)
        {
          return NULL;
        }

      ses->buflen += Bytes;

      if (ses->buflen == sizeof (ses->buffer) - 1)
        {
          return NULL;
        }
    }

#else
  return fgets (in, cap, ses->fpin);
#endif
}

static int
session_puts (session_t *ses, const char *out)
{
  int ret;
#ifdef WIN32
  WriteFile(ses->hWrite, out, strlen (out), &ret, NULL);
#else
  ret = fputs (out, ses->fpout);
  fflush (ses->fpout);
#endif
  return ret;
}
