/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE tport.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/07/24 17:15:29 Alf*/

#include "util.h"
#include "log.h"
#include "tport.h"

#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#else
#define socklen_t int
#endif

enum
{ AFE_READ, AFE_WRITE, AFE_ERROR };

static int fd_event (socket_t, int, int);

tport_t *
tport_new (int type, const char *ip, unsigned short port)
{
  socket_t fd;
  struct sockaddr_in loc[1];

  if (type == TCP || type == TLS)
    {
      fd = socket (AF_INET, SOCK_STREAM, 0);
    }
  else if (type == UDP)
    {
      fd = socket (AF_INET, SOCK_DGRAM, 0);
    }
  else
    {
      fd = -1;
    }

  if (fd < 0)
    {
      acc_debug (_("Socket error\n"));
      return NULL;
    }

  loc->sin_family = AF_INET;
  loc->sin_addr.s_addr = ip == NULL ? INADDR_ANY : inet_addr (ip);
  loc->sin_port = htons (port);
  memset (loc->sin_zero, 0, 8);
  if (bind (fd, (struct sockaddr *) loc, sizeof (struct sockaddr)) < 0)
    {
      acc_debug (_("Bind error\n"));
      close (fd);
      return NULL;
    }

  return tport_new_with_fd (type, fd);
}

tport_t *
tport_new_with_fd (int type, int fd)
{
  tport_t *tport;

  tport = (tport_t *) calloc (1, sizeof (tport_t));
  if (tport == NULL)
    {
      acc_debug (_("Memory error\n"));
      close (fd);
      return NULL;
    }

  tport->fd = fd;
  tport->type = type;

  return tport;
}

int
tport_connect (tport_t * tport, unsigned long ip, unsigned short port,
int timeout)
{
  int ret;
  struct sockaddr_in dst[1];

  if (socket_nonblock (tport->fd) < 0)
    {
      acc_debug (_("Set socket non-block error\n"));
      return -1;
    }

  dst->sin_family = AF_INET;
  dst->sin_addr.s_addr = ip;
  dst->sin_port = port;
  memset (dst->sin_zero, 0, 8);
  ret =
  connect (tport->fd, (struct sockaddr *) dst, sizeof (struct sockaddr));
  if (ret < 0)
    {
      if (errno != EINPROGRESS
      || fd_event (tport->fd, AFE_WRITE, timeout) <= 0)
        {
          acc_debug (_("Connect error\n"));
          return -1;
        }
    }

  if (socket_block (tport->fd) < 0)
    {
      acc_debug (_("Set socket block error\n"));
      return -1;
    }

  tport = (tport_t *) calloc (1, sizeof (tport_t));
  if (tport == NULL)
    {
      acc_debug (_("Memory error\n"));
      return -1;
    }

  *tport->rmt = *dst;

  return 0;
}

int
tport_write (tport_t * tport, const char *buf, int siz,
const struct sockaddr_in *sa)
{
  int ret;

  if (tport->fd < 0)
    {
      acc_debug ("fd error\n");
      return -1;
    }

  if (tport->type == TCP)
    {
      ret = write (tport->fd, buf, siz);
    }
  else if (tport->type == UDP)
    {
      ret =
      sendto (tport->fd, buf, siz, 0, (struct sockaddr *) sa,
      sizeof (struct sockaddr));
    }
  else
    {
      ret = -1;
    }
  /*  acc_debug ("socket %d send %d bytes to [%s:%d], return %d\n", tport->fd,
               siz, inet_ntoa (sa->sin_addr), ntohs (sa->sin_port), ret);*/

  if (ret < siz)
    {
      acc_debug (_("Write error\n"));
      return -1;
    }

  return 0;
}

int
tport_read (tport_t * tport, char *buf, int max, struct sockaddr_in *sa,
int timeout)
{
  int ret;
  socklen_t len;

  if (tport->fd < 0)
    {
      return -1;
    }

  ret = fd_event (tport->fd, AFE_READ, timeout);
  if (ret <= 0)
    {
      return -1;
    }

  if (tport->type == TCP)
    {
      ret = read (tport->fd, buf, max);
    }
  else if (tport->type == UDP)
    {
      len = sizeof (struct sockaddr);
      ret = recvfrom (tport->fd, buf, max, 0, (struct sockaddr *) sa, &len);
    }
  else
    {
      ret = -1;
    }

  return ret;
}

void
tport_delete (tport_t * tport)
{
  if (tport->fd > 0)
    {
      close (tport->fd);
    }
  free (tport);
}

static int
fd_event (socket_t fd, int event, int timeout)
{
  int ret;
  fd_set set[1];
  struct timeval tv[1];

  FD_ZERO (set);
  FD_SET (fd, set);
  tv->tv_sec = timeout / 1000;
  tv->tv_usec = (timeout % 1000) * 1000;

  switch (event)
    {
    case AFE_READ:
      ret = select ((int) (fd + 1), set, NULL, NULL, tv);
      break;

    case AFE_WRITE:
      ret = select ((int) (fd + 1), NULL, set, NULL, tv);
      break;

    case AFE_ERROR:
      ret = select ((int) (fd + 1), NULL, NULL, set, tv);
      break;

    default:
      return -1;
    }

  if (ret > 0 && FD_ISSET (fd, set))
    {
      return 1;
    }
  else if (ret == 0)
    {
      return 0;
    }
  else
    {
      acc_debug ("select error: %d(%s)\n", errno, strerror (errno));
      return -1;
    }
}

int
socket_block (int fd)
{
  int flag;

#ifndef WIN32
  flag = fcntl (fd, F_GETFL);
  if (fcntl (fd, F_SETFL, flag & ~O_NONBLOCK) < 0)
#else
  flag = 0;
  if (ioctlsocket (fd, FIONBIO, (unsigned long *) &flag) < 0)
#endif
    {
      return -1;
    }
  return 0;
}

int
socket_nonblock (int fd)
{
  int flag;

#ifndef WIN32
  flag = fcntl (fd, F_GETFL);
  if (fcntl (fd, F_SETFL, flag | O_NONBLOCK) < 0)
#else
  flag = 1;
  if (ioctlsocket (fd, FIONBIO, (unsigned long *) &flag) < 0)
#endif
    {
      return -1;
    }
  return 0;
}
