/*
* This file is part of the acnchess package
* Copyright (C) <2008>  <Alf>
*
* Contact: Alf <naihe2010@126.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with this program; if not, write to the Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*
*/
/* @CFILE util.c
*
*  Author: Alf <naihe2010@126.com>
*/
/* @date Created: 2009/08/11 13:58:37 Alf*/

#include "log.h"
#include "util.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#else
#include <winsock2.h>
#endif

static int
acc_is_realpath (const char *path)
{
#ifdef WIN32
  if (*path != '\0' && *(path + 1) == ':')
    {
      return 1;
    }
#else
  if (*path != '\0' && *path == '/')
    {
      return 1;
    }
#endif
  return 0;
}

char *
acc_realpath (const char *path)
{
  char abpath[PATH_MAX];

  if (acc_is_realpath (path) != 0)
    {
      return strdup (path);
    }

  if (*path == '~')
    {
#ifdef WIN32
      char *home =
        g_win32_get_package_installation_directory_of_module (NULL);
#else
      char *home = getenv ("HOME");
#endif
      snprintf (abpath, sizeof abpath, "%s%s", home, ++path);
    }
  else
    {
#ifdef WIN32
      getcwd (sizeof (abpath), abpath);
      strcat (abpath, "\\");
      strcat (abpath, path);
#else
      if (realpath (path, abpath) == NULL)
        {
          return NULL;
        }
#endif
    }

  return strdup (abpath);
}

int
str2url (const char *str, acc_url * url)
{
  char *h = (char *) str, *m1, *m2, *a, *x, *p, temp[PATH_MAX];
  size_t len;
  int ret;
  struct hostent *ent;

  ret = strncmp (h, "https://", 8);
  if (ret == 0)
    {
      url->type = HTTPS;
      url->port = HTTPS_PORT;
      h += 8;
    }
  else
    {
      ret = strncmp (h, "http://", 7);
      if (ret == 0)
        {
          url->type = HTTP;
          url->port = HTTP_PORT;
          h += 7;
        }
      else
        {
          acc_err (_("Parse URL error: Can't find 'http' from begin."));
          return -1;
        }
    }

  x = strchr (h, '/');

  a = strchr (h, '@');
  if (x && a > x)
    a = NULL;

  m1 = strchr (h, ':');
  if (x && m1 > x)
    m1 = NULL;

  if (a != NULL && (m1 == NULL || m1 > a))
    {
      acc_err (_("Parse URL error: Wrong username or password"));
      return -1;
    }

  if (a != NULL)
    m2 = strchr (a, ':');
  else
    m2 = m1;

  if (m2 != NULL)
    {
      if (x == NULL)
        url->port = (short) atoi (m2 + 1);
      else
        {
          len = x - m2;
          if (len <= 0)
            {
              acc_err (_("Parse URL error: Wrong port number"));
              return -1;
            }
          memcpy (temp, m2 + 1, len);
          temp[len] = 0;
          url->port = (short) atoi (temp);
        }
    }

  if (a != NULL)
    {
      len = m1 - h;
      memcpy (url->username, h, len);
      url->username[len] = 0;

      len = a - m1 - 1;
      memcpy (url->password, m1 + 1, len);
      url->password[len] = 0;

      h = a + 1;
    }

  if (m2 == NULL)
    {
      if (x != NULL)
        len = x - h;
      else
        len = strlen (h);
    }
  else
    {
      len = m2 - h;
    }
  memcpy (url->host, h, len);
  url->host[len] = 0;

  if (x != NULL)
    {
      p = strchr (x + 1, '/');
      if (p != NULL)
        {
          strcpy (url->doc, x);
        }
    }
  else
    {
      strcpy (url->doc, "/");
    }

  ent = gethostbyname (url->host);
  if (ent != NULL)
    {
      url->ip = *(unsigned long *) (ent->h_addr);
      snprintf (url->addr, sizeof url->addr, "%s",
                inet_ntoa (*(struct in_addr *) (ent->h_addr)));
    }
  else
    {
      return -1;
    }

  acc_debug ("url: %s\n", str);
  acc_debug ("url->host: %s\n", url->host);
  acc_debug ("url->username: %s\n", url->username);
  acc_debug ("url->password: %s\n", url->password);
  acc_debug ("url->doc: %s\n", url->doc);
  acc_debug ("url->type: %d\n", url->type);
  acc_debug ("url->addr: %s\n", url->addr);
  acc_debug ("url->port: %d\n", url->port);

  return 0;
}

static gint
right_button_press (GtkWidget * wid, GdkEvent * event)
{
  GdkEventButton *bevent;
  if (event->type == GDK_BUTTON_PRESS)
    {
      bevent = (GdkEventButton *) event;
      if (bevent->button == 3)
        {
          gtk_menu_popup (GTK_MENU (wid), NULL, NULL, NULL, NULL,
                          bevent->button, bevent->time);
          return TRUE;
        }
    }

  return FALSE;
}

void
gtk_add_popup_menu (GtkWidget * wid, GtkMenu * menu)
{
  g_signal_connect_swapped (G_OBJECT (wid), "event",
                            G_CALLBACK (right_button_press), menu);
}

void
gtk_menu_add_item (GtkWidget * menu, const char *label,
                   void (*func) (gpointer), gpointer data)
{
  GtkWidget *item;

  item = gtk_menu_item_new_with_mnemonic (label);
  gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
  g_signal_connect_swapped (G_OBJECT (item), "activate", G_CALLBACK (func),
                            data);
}
